# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List
from typing import Optional
from typing import Union
from unittest import TestCase

import numpy as np
import numpy.testing as npt
from parameterized.parameterized import parameterized

from kats.metrics import metrics


class MetricsTest(TestCase):
    def validate(self, expected: Union[float, np.ndarray], result: Union[float, np.ndarray]) -> None:
        if isinstance(expected, float):
            self.assertTrue(isinstance(result, float))
            if np.isnan(expected):
                self.assertTrue(np.isnan(result), f"{result} is not nan")
            else:
                self.assertAlmostEqual(expected, result)
        else:
            self.assertTrue(isinstance(result, np.ndarray))
            npt.assert_array_almost_equal(np.array(expected), result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", [[1.0, 2.0], [3.0, 4.0]], [[1.0, 2.0], [3.0, 4.0]]),
            ("missing", [[1, 2], [2, 3]], [[1, 2], None, [2, 3]]),
            # TODO: nans
            ("empty", [], []),
        ]
    )
    def test__arrays(self, _name: str, expected: List[List[float]], arrs: List[List[float]]) -> None:
        result = list(metrics._arrays(*arrs))
        self.assertEqual(len(expected), len(result))
        for e, a in zip(expected, result):
            npt.assert_almost_equal(np.array(e), a)

    def test__arrays_mismatched(self) -> None:
        with self.assertRaises(ValueError):
            _ = list(metrics._arrays([1], [2, 3]))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", [0.25, 0.5, 2.0], [1, 2, 4], [4, 4, 2]),
            # TODO: more numerical combinations
        ]
    )
    def test__safe_divide(
        self,
        _name: str,
        exp: List[float],
        num: List[float],
        denom: Union[float, List[float]],
        negative_infinity: float = -1.0,
        positive_infinity: float = 1.0,
        indeterminate: float = 0.0,
        nan: float = np.nan,
    ) -> None:
        expected = np.array(exp)
        numerator = np.array(num)
        denominator = np.array(denom)
        npt.assert_almost_equal(
            expected,
            metrics._safe_divide(
                numerator,
                denominator,
                negative_infinity,
                positive_infinity,
                indeterminate,
                nan,
            ),
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", [-3, -2, 2], [1, 2, 4], [4, 4, 2]),
            ("weighted", [-1.25, -0.125, 0.375], [1, 2, 3], [3.5, 2.5, 1.5], [2, 1, 1]),
            # TODO: more numerical combinations
            ("empty", [], [], []),
        ]
    )
    def test_error(
        self,
        _name: str,
        expected: List[float],
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.error(y_true, y_pred, weights)
        self.validate(np.array(expected), result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", [3, 2, 2], [1, 2, 4], [4, 4, 2]),
            ("weighted", [1.25, 0.125, 0.375], [1, 2, 3], [3.5, 2.5, 1.5], [2, 1, 1]),
            # TODO: more numerical combinations
            ("empty", [], [], []),
        ]
    )
    def test_absolute_error(
        self,
        _name: str,
        expected: List[float],
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.absolute_error(y_true, y_pred, weights)
        self.validate(np.array(expected), result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", [-3, -1, 0.5], [1, 2, 4], [4, 4, 2]),
            (
                "weighted",
                [-1.25, -0.0625, 0.125],
                [1, 2, 3],
                [3.5, 2.5, 1.5],
                [2, 1, 1],
            ),
            # TODO: more numerical combinations
            ("empty", [], [], []),
        ]
    )
    def test_percentage_error(
        self,
        _name: str,
        expected: List[float],
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.percentage_error(y_true, y_pred, weights)
        self.validate(np.array(expected), result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", [3, 1, 0.5], [1, 2, 4], [4, 4, 2]),
            ("weighted", [1.25, 0.0625, 0.125], [1, 2, 3], [3.5, 2.5, 1.5], [2, 1, 1]),
            # TODO: more numerical combinations
            ("empty", [], [], []),
        ]
    )
    def test_absolute_percentage_error(
        self,
        _name: str,
        expected: List[float],
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.absolute_percentage_error(y_true, y_pred, weights)
        self.validate(np.array(expected), result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 2 / 9, [1, 2, 4], [4, 4, 2]),
            # TODO: more numerical combinations
            ("all_pred_missing", np.nan, [1.0, 2.0], [np.nan, np.nan]),
            ("all_true_missing", 1.0, [np.nan, np.nan], [1.0, 2.0]),
            ("empty", np.nan, [], []),
        ]
    )
    def test_continuous_rank_probability_score(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.crps(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 0.5, [2, 3, 4, -5], [1, 1, 1, 1], 3.0),
            ("empty", np.nan, [], [], 1.0),
        ]
    )
    def test_frequency_exceeds_relative_threshold(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        threshold: float,
    ) -> None:
        result = metrics.frequency_exceeds_relative_threshold(y_true, y_pred, threshold)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 4 / 9, [1, 2, 4], [4, 4, 2]),
            # TODO: more numerical combinations
            ("all_pred_missing", np.nan, [1.0, 2.0], [np.nan, np.nan]),
            ("all_true_missing", 1.0, [np.nan, np.nan], [1.0, 2.0]),
            ("empty", np.nan, [], []),
        ]
    )
    def test_linear_error_in_probability_space(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.leps(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 1.5, [1, 2, 3], [3.5, 2.5, 1.5]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_median_absolute_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.mdae(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 0.5, [1, 2, 3], [3.5, 2.5, 1.5]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_median_absolute_percent_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.mdape(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 1.5, [1, 2, 3], [3.5, 2.5, 1.5]),
            ("weighted", 1.75, [1, 2, 3], [3.5, 2.5, 1.5], [2, 1, 1]),
            ("normal_2d", 2.25, [[1, 2, 3], [2, 4, 6]], [[3.5, 2.5, 1.5], [1, 1, 1]]),
            (
                "normal_2d_raw",
                [1.75, 1.75, 3.25],
                [[1, 2, 3], [2, 4, 6]],
                [[3.5, 2.5, 1.5], [1, 1, 1]],
                None,
                "raw_values",
            ),
            (
                "weighted_2d",
                2.458333333333,
                [[1, 2, 3], [2, 4, 6]],
                [[3.5, 2.5, 1.5], [1, 1, 1]],
                [[1, 1, 1], [1, 3, 1]],
            ),
            (
                "weighted_2d_raw",
                [1.75, 2.375, 3.25],
                [[1, 2, 3], [2, 4, 6]],
                [[3.5, 2.5, 1.5], [1, 1, 1]],
                [[1, 1, 1], [1, 3, 1]],
                "raw_values",
            ),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_mean_absolute_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
        multioutput: Union[str, List[float]] = "uniform_average",
    ) -> None:
        result = metrics.mae(y_true, y_pred, weights, multioutput)
        self.validate(expected, result)

    def test_invalid_mean_absolute_error_multioutput(self) -> None:
        with self.assertRaises(ValueError):
            _ = metrics.mae([], [], [], "mango")

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 2.125 / 2, [-2, 2, 3, 4], [0.5, 2.5, 1.5, 0]),
            # TODO: more numerical combinations
            ("zero_baseline", np.nan, [1, 1, 1, 1], [3.5, 2.5, 1.5, 0]),
            ("empty", np.nan, [], []),
        ]
    )
    def test_mean_absolute_scaled_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.mase(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 1.0625, [1, 2, 3, 4], [3.5, 2.5, 1.5, 0]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_mean_absolute_percent_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.mape(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 0.625, [1, 2, 3, 4], [3.5, 2.5, 1.5, 0]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_mean_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.me(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", -0.3125, [1, 2, 3, 4], [3.5, 2.5, 1.5, 0]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_mean_percentage_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.mpe(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 35 / 12, [1, 2, 3], [3.5, 2.5, 1.5]),
            ("weighted", 2.75, [1, 2, 3], [3.5, 2.5, 1.5], [1, 1, 2]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_mean_squared_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.mse(y_true, y_pred, weights)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", np.sqrt(35 / 12), [1, 2, 3], [3.5, 2.5, 1.5]),
            ("weighted", np.sqrt(2.75), [1, 2, 3], [3.5, 2.5, 1.5], [1, 1, 2]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_root_mean_squared_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.rmse(y_true, y_pred, weights)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 0.5484140, [1, 2, 3], [3.5, 2.5, 1.5]),
            ("weighted", 0.5299002, [1, 2, 3], [3.5, 2.5, 1.5], [1, 1, 2]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_root_mean_squared_log_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.rmsle(y_true, y_pred, weights)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 1.4790199, [1, 2, 3], [3.5, 2.5, 1.5]),
            ("weighted", 0.3903124, [1, 2, 3], [3.5, 2.5, 1.5], [1, 1, 2]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_root_mean_squared_percentage_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        weights: Optional[List[float]] = None,
    ) -> None:
        result = metrics.rmspe(y_true, y_pred, weights)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 1 / 3, [1, 2, 3], [3.5, 2.5, 1.5]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_scaled_symmetric_mean_absolute_percentage_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.scaled_smape(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 2 / 9, [1, 2, 3], [3.5, 2.5, 1.5]),
            ("negatives", 2 / 27, [-1, 2, 3], [3.5, 2.5, -1.5]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_symmetric_bias(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.sbias(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 2 / 3, [1, 2, 3], [3.5, 2.5, 1.5]),
            (
                "normal",
                0.673220686123912,
                [0.2, 0.2, 0.2, 0.8, 0.8],
                [0.0, 0.25, 0.5, 0.75, 1.0],
            ),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_symmetric_mean_absolute_percentage_error(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.smape(y_true, y_pred)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", -1.0, [1, 2, 3], [3.5, 2.5, 1.5]),
            # TODO: more numerical combinations
            ("empty", np.nan, [], []),
        ]
    )
    def test_tracking_signal(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
    ) -> None:
        result = metrics.tracking_signal(y_true, y_pred)
        self.validate(expected, result)

    def test_metric(self) -> None:
        self.assertEqual(metrics.mae, metrics.metric("mae"))
        with self.assertRaises(ValueError):
            _ = metrics.metric("mango")

    def test_core_metric(self) -> None:
        self.assertEqual(metrics.mae, metrics.core_metric("mae"))
        with self.assertRaises(ValueError):
            _ = metrics.metric("mango")

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            (
                "normal",
                [0.4, 0.2, 0.33333333],
                [
                    (0.40547657, 0.21700191, -0.63343906, 0.24662161, -1.9395454),
                    (-0.04428768, 0.5543952, -0.40847492, -0.46409416, 0.4180088),
                    (-2.0893543, -0.12981987, -0.58653784, -0.58653784, 0.29072),
                ],
                [
                    (
                        -1.5253627,
                        -2.0157309,
                        -1.3632555,
                        1.8552899,
                        5.08259,
                        8.782536,
                        4.62253,
                        -0.73543787,
                        2.656838,
                        2.5200548,
                        9.273176,
                        2.6641555,
                        -0.39546585,
                        0.5721655,
                        -1.0635448,
                    ),
                    (
                        -3.8493829,
                        -3.2209146,
                        -2.5079165,
                        -1.3597498,
                        4.16947,
                        3.6076689,
                        3.6549635,
                        -1.8097634,
                        -0.76120234,
                        1.5070448,
                        4.0525684,
                        1.6184692,
                        -1.4960217,
                        -2.3242073,
                        -2.226036,
                    ),
                    (
                        -2.127775,
                        -2.4119477,
                        -0.58012056,
                        0.4478078,
                        3.292698,
                        5.592966,
                        2.9125519,
                        0.27569342,
                        1.0328965,
                        1.2424107,
                        6.086138,
                        1.2846599,
                        0.6023383,
                        -0.61473894,
                        -1.641422,
                    ),
                ],
                [0.05, 0.95, 0.99],
            ),
        ]
    )
    def test_mult_exceed(
        self,
        _name: str,
        expected: List[float],
        y_true: List[float],
        y_pred: List[float],
        threshold: List[float],
    ) -> None:
        result = metrics.mult_exceed(y_true, y_pred, threshold)
        self.validate(np.array(expected), result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 1.3333333, [1, 5, 1], [5, 1, 1], 0),
            ("normal", 1.3333333, [1, 5, 1], [5, 1, 1], 0.5),
            ("normal", 1.3333333, [1, 5, 1], [5, 1, 1], 0.75),
            ("normal", 1.3333333, [1, 5, 1], [5, 1, 1], 1),
            # # TODO: more numerical combinations
            ("empty", np.nan, [], [], 1.0),
        ]
    )
    def test_pinball_loss(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        threshold: float,
    ) -> None:
        result = metrics.pinball_loss(y_true, y_pred, threshold)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            (
                "normal",
                0.6,
                [
                    (0.40547657, 0.21700191, -0.63343906, 0.24662161, -1.9395454),
                    (-0.04428768, 0.5543952, -0.40847492, -0.46409416, 0.4180088),
                    (-2.0893543, -0.12981987, -0.58653784, -0.58653784, 0.29072),
                ],
                [
                    (-1.5253627, -2.0157309, -1.3632555, 1.8552899, 5.08259),
                    (-3.8493829, -3.2209146, -2.5079165, -1.3597498, 4.16947),
                    (-2.127775, -2.4119477, -0.58012056, 0.4478078, 3.292698),
                ],
                0.95,
            ),
            ("normal", 0.67, [1, 2, 3], [3.5, 2.5, 1.5], 0.25),
            ("normal", 0.67, [1, 2, 3], [3.5, 2.5, 1.5], 0.50),
            ("normal", 0.33, [1, 2, 3], [3.5, 2.5, 1.5], 0.75),
            ("empty", np.nan, [], [], 1.0),
        ]
    )
    def test_exceed(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_pred: List[float],
        threshold: float,
    ) -> None:
        result = metrics.exceed(y_true, y_pred, threshold)
        self.validate(expected, round(result, 2))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 0.33, [1.1, 2.0, 3.31], [1.2, 2.4, 3.3], [1.5, 2.2, 3.4]),
            ("empty", np.nan, [], [], []),
        ]
    )
    def test_coverage(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_lower: List[float],
        y_upper: List[float],
    ) -> None:
        result = metrics.coverage(y_true, y_lower, y_upper)
        self.validate(expected, round(result, 2))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            (
                "normal",
                [0, 0, 1],
                [1.1, 2.0, 3.31],
                [1.2, 2.4, 3.3],
                [1.5, 2.2, 3.4],
                None,
            ),
            ("empty", [], [], [], [], None),
        ]
    )
    def test_mult_coverage(
        self,
        _name: str,
        expected: np.ndarray,
        y_true: List[float],
        y_lower: List[float],
        y_upper: List[float],
        rolling_window: Union[None, int],
    ) -> None:
        result = metrics.mult_coverage(y_true, y_lower, y_upper, rolling_window)
        self.validate(expected, result)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            ("normal", 1.73, [1.1, 2.0, 3.31], [1.2, 2.4, 3.3], [1.5, 2.2, 3.4], 0.2),
            ("empty", np.nan, [], [], [], 0.2),
        ]
    )
    def interval_score(
        self,
        _name: str,
        expected: float,
        y_true: List[float],
        y_lower: List[float],
        y_upper: List[float],
        alpha: float,
    ) -> None:
        result = metrics.interval_score(y_true, y_lower, y_upper)
        self.validate(expected, round(result, 2))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator...
    @parameterized.expand(
        [
            (
                "normal",
                [1.3, 3.8, 0.1],
                [1.1, 2.0, 3.31],
                [1.2, 2.4, 3.3],
                [1.5, 2.2, 3.4],
                0.2,
                None,
            ),
            ("empty", [], [], [], [], 0.2, None),
        ]
    )
    def test_mult_interval_score(
        self,
        _name: str,
        expected: np.ndarray,
        y_true: List[float],
        y_lower: List[float],
        y_upper: List[float],
        alpha: float,
        rolling_window: Union[None, int],
    ) -> None:
        result = metrics.mult_interval_score(y_true, y_lower, y_upper, alpha, rolling_window)
        self.validate(expected, result)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Median ensembling method

Ensemble models with median of individual models
Assume we have k base models, after we make forecasts with each individual
model, we take the median from each time point as the final results
"""

import logging
from typing import List
from typing import Optional
from typing import cast

import pandas as pd

from kats.consts import TimeSeriesData
from kats.models.ensemble import ensemble
from kats.models.ensemble.ensemble import EnsembleParams


class MedianEnsembleModel(ensemble.BaseEnsemble):
    """Median ensemble model class

    Attributes:
        data: the input time series data as in :class:`kats.consts.TimeSeriesData`
        params: the model parameter class in Kats
    """

    freq: Optional[str] = None
    fcst: Optional[pd.DataFrame] = None
    fcst_dates: Optional[pd.DatetimeIndex] = None
    dates: Optional[pd.DatetimeIndex] = None
    fcst_df: Optional[pd.DataFrame] = None

    def __init__(self, data: TimeSeriesData, params: EnsembleParams) -> None:
        self.data = data
        self.params = params
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, **kwargs) -> pd.DataFrame:
        """Predict method of median ensemble model

        Args:
            steps: the length of forecasting horizon

        Returns:
            forecasting results as in pd.DataFrame
        """

        logging.debug(
            "Call predict() with parameters. " "steps:{steps}, kwargs:{kwargs}".format(steps=steps, kwargs=kwargs)
        )
        # Keep freq in the parameters passed to _predict_all()
        self.freq = freq = kwargs.get("freq", "D")
        pred_dict = self._predict_all(steps, **kwargs)

        fcst_all = pd.concat(
            # pyre-fixme[16]: `Model` has no attribute `fcst`.
            [x.fcst.reset_index(drop=True) for x in pred_dict.values()],
            axis=1,
            copy=False,
        )
        fcst_all.columns = cast(List[str], pred_dict.keys())
        self.fcst = fcst_all.median(axis=1)

        # create future dates
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=freq)
        dates = dates[dates != last_date]
        self.fcst_dates = dates.to_pydatetime()
        self.dates = dates[dates != last_date]
        self.fcst_df = fcst_df = pd.DataFrame({"time": self.dates, "fcst": self.fcst}, copy=False)

        logging.debug("Return forecast data: {fcst_df}".format(fcst_df=self.fcst_df))
        return fcst_df

    def __str__(self) -> str:
        """Get default parameter search space for the median ensemble model

        Args:
            None

        Returns:
            Model name as a string
        """
        return "Median Ensemble"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from operator import attrgetter
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.detectors.outlier import MultivariateAnomalyDetector
from kats.detectors.outlier import MultivariateAnomalyDetectorType
from kats.detectors.outlier import OutlierDetector
from kats.models.bayesian_var import BayesianVARParams
from kats.models.var import VARParams


# Anomaly detection tests
class OutlierDetectionTest(TestCase):
    def setUp(self) -> None:
        data = load_air_passengers(return_ts=False)
        self.ts_data = TimeSeriesData(data)
        data_2 = data.copy()
        data_2["y_2"] = data_2["y"]
        self.ts_data_2 = TimeSeriesData(data_2)

        daily_data = load_data("peyton_manning.csv")
        daily_data.columns = ["time", "y"]
        self.ts_data_daily = TimeSeriesData(daily_data)

        daily_data_missing = daily_data.drop([2, 11, 18, 19, 20, 21, 22, 40, 77, 101]).copy()
        # Detecting missing data is coupled to pd.infer_freq() implementation. Make sure the
        # rows dropped above prevent us from inferring a frequency (so it returns None)
        self.assertIsNone(pd.infer_freq(daily_data_missing["time"]))
        self.ts_data_daily_missing = TimeSeriesData(daily_data_missing)

    def test_additive_overrides_missing_daily_data(self) -> None:
        m = OutlierDetector(self.ts_data_daily_missing, "additive")

        m.detector()
        outliers = m.remover(interpolate=True)

        m2 = OutlierDetector(self.ts_data_daily_missing, "logarithmic")

        m2.detector()
        outliers2 = m2.remover(interpolate=True)

        self.assertEqual(outliers.value.all(), outliers2.value.all())

    def test_additive_overrides(self) -> None:
        m = OutlierDetector(self.ts_data, "additive")

        m.detector()
        outliers = m.remover(interpolate=True)

        m2 = OutlierDetector(self.ts_data, "logarithmic")

        m2.detector()
        outliers2 = m2.remover(interpolate=True)

        self.assertEqual(outliers.value.all(), outliers2.value.all())

    def test_outlier_detection_additive(self) -> None:
        m = OutlierDetector(self.ts_data, "additive")

        m.detector()
        m.remover(interpolate=True)

        m2 = OutlierDetector(self.ts_data_daily, "additive")
        m2.detector()
        m2.remover(interpolate=True)
        # test for multiple time series
        m3 = OutlierDetector(self.ts_data_2, "additive")
        m3.detector()
        m3.remover(interpolate=True)

    def test_outlier_detection_multiplicative(self) -> None:
        m = OutlierDetector(self.ts_data, "multiplicative")
        m.detector()
        m.remover(interpolate=True)

        m2 = OutlierDetector(self.ts_data_daily, "multiplicative")
        m2.detector()
        m2.remover(interpolate=True)
        # test for multiple time series
        m3 = OutlierDetector(self.ts_data_2, "additive")
        m3.detector()
        m3.remover(interpolate=True)

    def test_outlier_detector_exception(self) -> None:
        data = self.ts_data.to_dataframe()
        data_new = pd.concat([data, data])
        ts_data_new = TimeSeriesData(data_new)

        with self.assertLogs(level="ERROR"):
            m = OutlierDetector(ts_data_new)
            m.detector()

    def test_output_scores_not_nan(self) -> None:
        m = OutlierDetector(self.ts_data, "additive")
        m.detector()
        output_scores = m.output_scores

        # test that there are not any nan values in output score
        self.assertIsNotNone(output_scores)
        if output_scores is not None:
            self.assertFalse(output_scores["y"].isnull().any())

    def test_output_scores_shape_one_series(self) -> None:
        # test for a single time series
        m = OutlierDetector(self.ts_data, "additive")
        m.detector()
        output_scores = m.output_scores

        self.assertIsNotNone(output_scores)
        if output_scores is not None:
            self.assertEqual(output_scores.shape[0], m.data.value.shape[0])

    def test_output_scores_shape_two_series(self) -> None:
        # test for more than 1 time series
        m2 = OutlierDetector(self.ts_data_2, "additive")
        m2.detector()
        output_scores2 = m2.output_scores

        self.assertIsNotNone(output_scores2)
        if output_scores2 is not None:
            self.assertEqual(output_scores2.shape, m2.data.value.shape)


class MultivariateVARDetectorTest(TestCase):
    def setUp(self) -> None:
        DATA_multi = load_data("multivariate_anomaly_simulated_data.csv")
        self.TSData_multi = TimeSeriesData(DATA_multi)
        DATA_multi2 = pd.concat([DATA_multi, DATA_multi])
        self.TSData_multi2 = TimeSeriesData(DATA_multi2)
        DATA_multi3 = pd.merge(DATA_multi, DATA_multi, how="inner", on="time", suffixes=("_1", "_2"))
        self.TSData_multi3 = TimeSeriesData(DATA_multi3)

        self.params_var = VARParams(maxlags=2)
        self.d_var = MultivariateAnomalyDetector(self.TSData_multi, self.params_var, training_days=60)

        self.params_bayes = BayesianVARParams(p=2)
        self.d_bayes = MultivariateAnomalyDetector(
            self.TSData_multi,
            self.params_bayes,
            training_days=60,
            model_type=MultivariateAnomalyDetectorType.BAYESIAN_VAR,
        )

    # pyre-ignore Undefined attribute [16]: Module parameterized.parameterized has no attribute expand.
    @parameterized.expand(
        [
            ["d_var"],
            ["d_bayes"],
        ]
    )
    def test_detector_count_equal(self, attribute: str) -> None:
        np.random.seed(10)

        d = attrgetter(attribute)(self)
        anomaly_score_df = d.detector()
        self.assertCountEqual(
            list(anomaly_score_df.columns),
            list(self.TSData_multi.value.columns) + ["overall_anomaly_score", "p_value"],
        )

    # pyre-ignore Undefined attribute [16]: Module parameterized.parameterized has no attribute expand.
    @parameterized.expand(
        [
            ["d_var"],
            ["d_bayes"],
        ]
    )
    def test_detector_plot(self, attribute: str) -> None:
        np.random.seed(10)

        d = attrgetter(attribute)(self)
        d.detector()
        d.plot()

    # pyre-ignore Undefined attribute [16]: Module parameterized.parameterized has no attribute expand.
    @parameterized.expand(
        [
            ["d_var"],
            ["d_bayes"],
        ]
    )
    def test_detector_get_anomalous_metrics(self, attribute: str) -> None:
        np.random.seed(10)

        d = attrgetter(attribute)(self)
        d.detector()
        alpha = 0.05
        anomalies = d.get_anomaly_timepoints(alpha)
        d.get_anomalous_metrics(anomalies[0], top_k=3)

    # pyre-ignore Undefined attribute [16]: Module parameterized.parameterized has no attribute expand.
    @parameterized.expand(
        [
            ["TSData_multi2"],
            ["TSData_multi3"],
        ]
    )
    def test_runtime_errors(self, attribute: str) -> None:
        with self.assertRaises(RuntimeError):
            d = MultivariateAnomalyDetector(attrgetter(attribute)(self), self.params_var, training_days=60)
            d.detector()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any
from typing import Generic
from typing import Optional
from typing import Tuple
from typing import TypeVar

import matplotlib.pyplot as plt
import pandas as pd

from kats.consts import TimeSeriesData

ParamsType = TypeVar("ParamsType")


class Model(Generic[ParamsType]):
    __slots__ = ["data"]

    """Base forecasting model

    This is the parent class for all forecasting models in Kats

    Attributes:
        data: `TimeSeriesData` object
        params: model parameters
        validate_frequency: validate the frequency of time series
        validate_dimension: validate the dimension of time series
    """
    data: Optional[TimeSeriesData]
    fcst_df: Optional[pd.DataFrame] = None
    include_history: bool = False

    def __init__(
        self,
        data: Optional[TimeSeriesData],
        params: ParamsType,
        validate_frequency: bool = False,
        validate_dimension: bool = False,
    ) -> None:
        self.data = data
        self.params = params
        self.__type__ = "model"
        if data is not None:
            data.validate_data(validate_frequency, validate_dimension)

    def setup_data(self) -> None:
        """abstract method to set up dataset

        This is a declaration for setup data method
        """
        pass

    def validate_inputs(self) -> None:
        """abstract method to validate the inputs

        This is a declaration for validate_inputs method
        """
        pass

    def fit(self) -> None:
        """abstract method to fit model

        This is a declaration for model fitting
        """
        pass

    def predict(self, *_args: Any, **_kwargs: Any) -> None:
        """abstract method to predict

        This is a declaration for predict method
        """
        pass

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[int, int]] = None,
        **kwargs: Any,
    ) -> plt.Axes:
        """Plot method for forecasting models

        This method provides base plotting functionality for all forecasting
        models.

        Args:
            ax: optional Matplotlib Axes to use.
        Returns:
            The matplotlib Axes object.
        """
        fcst_df = self.fcst_df
        if fcst_df is None:
            raise ValueError("predict() must be called before plot().")
        data = self.data
        assert data is not None
        include_history = self.include_history

        logging.info("Generating chart for forecast result.")
        if ax is None:
            if figsize is None:
                figsize = (10, 6)
            fig, ax = plt.subplots(facecolor="w", figsize=figsize)
        else:
            fig = plt.gcf()

        # Allow subclasses to specify different kwargs by fetching them
        # here instead of in the method signature.
        intervals = kwargs.get("intervals", True)
        ls = kwargs.get("ls", "-")
        history_color = kwargs.get("history_color", "k")
        forecast_color = kwargs.get("forecast_color", "#4267B2")
        grid = kwargs.get("grid", True)
        xlabel = kwargs.get("xlabel", "time")
        ylabel = kwargs.get("ylabel", "y")

        if grid:
            ax.grid(True, which="major", c="gray", ls="-", lw=1, alpha=0.2)

        if include_history:
            ax.plot(data.time, data.value, history_color)

        ax.plot(fcst_df["time"], fcst_df["fcst"], ls=ls, c=forecast_color)

        if intervals and {"fcst_lower", "fcst_upper"}.issubset(fcst_df.columns):
            ax.fill_between(
                fcst_df["time"],
                fcst_df["fcst_lower"],
                fcst_df["fcst_upper"],
                color=forecast_color,
                alpha=0.2,
            )

        ax.set_xlabel(xlabel=xlabel)
        ax.set_ylabel(ylabel=ylabel)
        fig.set_tight_layout(True)
        return ax

    @staticmethod
    def get_parameter_search_space() -> None:
        """method to query default parameter search space

        abstract method to be implemented by downstream forecasting models
        """
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# This file defines tests for the Backtester classes

import statistics
import unittest
import unittest.mock as mock
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import cast

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.metrics.metrics import core_metric
from kats.tests.test_backtester_dummy_data import PROPHET_0_72_FCST_DUMMY_DATA
from kats.tests.test_backtester_dummy_data import PROPHET_0_72_GAP_36_FCST_DUMMY_DATA
from kats.tests.test_backtester_dummy_data import PROPHET_0_90_FCST_DUMMY_DATA
from kats.tests.test_backtester_dummy_data import PROPHET_0_108_FCST_DUMMY_DATA
from kats.tests.test_backtester_dummy_data import PROPHET_18_90_FCST_DUMMY_DATA
from kats.tests.test_backtester_dummy_data import PROPHET_36_108_FCST_DUMMY_DATA
from kats.tests.test_backtester_dummy_data import PROPHET_EMPTY_DUMMY_DATA
from kats.utils.backtesters import BackTesterExpandingWindow
from kats.utils.backtesters import BackTesterFixedWindow
from kats.utils.backtesters import BacktesterResult
from kats.utils.backtesters import BackTesterRollingWindow
from kats.utils.backtesters import BackTesterSimple
from kats.utils.backtesters import CrossValidation
from kats.utils.backtesters import KatsSimpleBacktester
from kats.utils.backtesters import _return_fold_offsets as return_fold_offsets
from kats.utils.datapartition import SimpleDataPartition

# Constants
ALL_ERRORS = ["mape", "smape", "mae", "mase", "mse", "rmse"]  # Errors to test
TIMESTEPS = 36  # Timesteps for test data
FREQUENCY = "MS"  # Frequency for model
PERCENTAGE = 75  # Percentage of train data
EXPANDING_WINDOW_START = 50  # Expanding window start training percentage
EXPANDING_WINDOW_STEPS = 3  # Expanding window number of steps
ROLLING_WINDOW_TRAIN = 50  # Rolling window start training percentage
ROLLING_WINDOW_STEPS = 3  # Rolling window number of steps
FIXED_WINDOW_TRAIN_PERCENTAGE = 50  # Fixed window ahead training percentage
FIXED_WINDOW_PERCENTAGE = 25  # Fixed window ahead window percentage
FLOAT_ROUNDING_PARAM = 3  # Number of decimal places to round low floats to 0
CV_NUM_FOLDS = 3  # Number of folds for cross validation


def compute_errors(train: np.ndarray, pred: np.ndarray, truth: np.ndarray) -> Dict[str, float]:
    true_errors = {}
    for error in ALL_ERRORS:
        func = core_metric(error)
        err = func(truth, pred)
        true_errors[error] = err
    return true_errors


def compute_errors_list(
    train: np.ndarray,
    pred: np.ndarray,
    truth: np.ndarray,
    true_errors: Dict[str, List[float]],
) -> None:
    for error in ALL_ERRORS:
        func = core_metric(error)
        err = func(truth, pred)
        if error in true_errors:
            true_errors[error].append(err)
        else:
            true_errors[error] = [err]


class SimpleBackTesterTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Setting up data
        DATA = load_air_passengers(return_ts=False)
        cls.TSData = load_air_passengers()
        cls.train_data = cls.TSData[: len(cls.TSData) - TIMESTEPS]
        # pyre-fixme[6]: For 1st param expected `Optional[DataFrame]` but got
        #  `Union[DataFrame, Series]`.
        cls.test_data = TimeSeriesData(DATA.tail(TIMESTEPS))

    def prophet_predict_side_effect(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        if len(kwargs) > 1 and kwargs["steps"] == TIMESTEPS and kwargs["freq"] == FREQUENCY:
            return PROPHET_0_108_FCST_DUMMY_DATA
        else:
            return PROPHET_EMPTY_DUMMY_DATA

    def test_error_values(self) -> None:
        """
        Testing process consists of the following:
          1. Backtest with mocked model
          2. Ensure backtester used correct train test splits
          3. Train (mocked) model locally and extract errors
          4. Comparing errors to backtester results
        """

        # Mock model results
        model_class = mock.MagicMock()
        model_class().predict.side_effect = self.prophet_predict_side_effect
        model_params = mock.MagicMock()

        # Creating and running backtester for this given model class
        bt = BackTesterSimple(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=model_params,
            train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            model_class=model_class,
        )
        bt.run_backtest()

        # Testing that backtester initialized model with correct train/test split
        model_class.assert_called_with(data=self.train_data, params=model_params)

        # Training local model and getting predictions
        temp_model = model_class(data=self.train_data, params=model_params)
        temp_model.fit()
        temp_fcst = temp_model.predict(steps=TIMESTEPS, freq=FREQUENCY)

        # Using model predictions from local model to calculate true errors
        pred = np.array(temp_fcst["fcst"])
        truth = np.array(self.test_data.value)
        train = np.array(self.train_data.value)
        true_errors = compute_errors(train, pred, truth)
        ground_truth_errors = (bt, true_errors)

        # Comparing local model errors to backtester errors
        backtester, error_results = ground_truth_errors
        for error_name, value in error_results.items():
            self.assertEqual(
                round(value, FLOAT_ROUNDING_PARAM),
                round(backtester.errors[error_name], FLOAT_ROUNDING_PARAM),
            )

    def test_forbidden_train_test_splits(self) -> None:
        """
        Testing process consists of the following:
          1. Backtest with mocked model
          2. Set forbidden train test split values
          3. Ensure backtester raises an exception
        """

        # Mock model results
        model_class = mock.MagicMock()
        model_class().predict.side_effect = self.prophet_predict_side_effect
        model_params = mock.MagicMock()

        # Create backtester for this given model class
        bt = BackTesterSimple(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=model_params,
            train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            model_class=model_class,
        )

        # Create forbidden train test split values
        forbidden_train_test_splits = [
            (-10, 50, "Incorrect training size"),
            (50, -10, "Incorrect testing size"),
            (60, 60, "Incorrect training and testing sizes"),
        ]

        for train_p, test_p, expected_msg in forbidden_train_test_splits:
            # Set backtester with forbidden train test split values
            bt.train_percentage = train_p
            bt.test_percentage = test_p

            # Ensure backtester will raise an expection
            with self.assertRaises(ValueError) as e:
                bt._create_train_test_splits()

            generated_msg = str(e.exception)
            self.assertEqual(expected_msg, generated_msg)


class ExpandingWindowBackTesterTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Setting up data
        DATA = load_air_passengers(return_ts=False)
        cls.TSData = load_air_passengers()

        cls.train_folds, cls.test_folds = cls.create_folds(
            data=DATA,
            start_train_size=EXPANDING_WINDOW_START / 100.0 * len(DATA),
            end_train_size=PERCENTAGE / 100.0 * len(DATA),
            test_size=(100 - PERCENTAGE) / 100.0 * len(DATA),
            num_folds=EXPANDING_WINDOW_STEPS,
        )

        # Mock model results
        cls.model_class = mock.MagicMock()
        cls.model_class().predict.side_effect = [
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_0_90_FCST_DUMMY_DATA,
            PROPHET_0_108_FCST_DUMMY_DATA,
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_0_90_FCST_DUMMY_DATA,
            PROPHET_0_108_FCST_DUMMY_DATA,
        ]  # Once for backtester once for local model
        cls.model_params = mock.MagicMock()

    def test_error_values(self) -> None:
        """
        Testing process consists of the following:
          1. Backtest with mocked model
          2. Ensure backtester used correct train test splits
          3. Train (mocked) model locally and extract errors
          4. Comparing errors to backtester results
        """

        # Create and run backtester
        bt = BackTesterExpandingWindow(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            start_train_percentage=EXPANDING_WINDOW_START,
            end_train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            expanding_steps=EXPANDING_WINDOW_STEPS,
            model_class=self.model_class,
            multi=False,
        )
        bt.run_backtest()

        # Test model initialization
        self.assertEqual(self.model_class.call_count, 4)
        self.model_class.assert_has_calls(
            [
                mock.call(data=TimeSeriesData(self.train_folds[0]), params=self.model_params),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(data=TimeSeriesData(self.train_folds[1]), params=self.model_params),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(data=TimeSeriesData(self.train_folds[2]), params=self.model_params),
            ]
        )

        # Test model predict
        # self.assertEqual(self.model_class().predict.call_count, 3)
        self.model_class().assert_has_calls(
            [
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
            ]
        )

        # Training local model and getting predictions
        true_errors = {}  # Dict to store errors
        for i in range(0, len(self.train_folds)):
            train_fold = self.train_folds[i]
            test_fold = self.test_folds[i]
            temp_model = self.model_class(data=TimeSeriesData(train_fold), params=self.model_params)
            temp_model.fit()

            # Getting model predictions
            temp_fcst = temp_model.predict(steps=TIMESTEPS, freq=FREQUENCY)

            # Using model predictions from local_model to calculate true errors
            pred = np.array(temp_fcst["fcst"])
            truth = np.array(test_fold["y"])
            train = np.array(train_fold["y"])
            compute_errors_list(train, pred, truth, true_errors)

        # Calculating errors
        for error_name, values in true_errors.items():
            true_errors[error_name] = statistics.mean(values)
        ground_truth_errors = (bt, true_errors)

        # # Test that local model errors equal backtester errors
        backtester, error_results = ground_truth_errors
        for error_name, value in error_results.items():
            self.assertEqual(
                round(value, FLOAT_ROUNDING_PARAM),
                round(backtester.errors[error_name], FLOAT_ROUNDING_PARAM),
            )

    def test_one_step_forecast(self) -> None:
        """
        Tests that if expanding steps is one, the folds returned are the same
        as BacktesterSimple.
        """

        # Calculate folds from Expanding Window Backtester
        expanding_backtester = BackTesterExpandingWindow(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            start_train_percentage=PERCENTAGE,
            end_train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            expanding_steps=1,
            model_class=self.model_class,
            multi=False,
        )
        one_step_folds_expanding = expanding_backtester._create_train_test_splits()

        # Calculate folds from Simple Backtester
        simple_backtester = BackTesterSimple(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            model_class=self.model_class,
        )
        folds_simple = simple_backtester._create_train_test_splits()

        # Test that folds are equivalent
        self.assertEqual(one_step_folds_expanding, folds_simple)

    def test_forbidden_initialization_parameters(self) -> None:
        """
        Testing process consists of the following:
          1. Create a Backtest instance with forbidden initial parameters
          2. Ensure backtester raises an exception
        """

        forbidden_init_params = [
            (-10, 50, EXPANDING_WINDOW_STEPS, "Invalid start training percentage"),
            (110, 50, EXPANDING_WINDOW_STEPS, "Invalid end training percentage"),
            (50, -10, EXPANDING_WINDOW_STEPS, "Invalid test percentage"),
            (50, 110, EXPANDING_WINDOW_STEPS, "Invalid test percentage"),
            (
                60,
                60,
                EXPANDING_WINDOW_STEPS,
                "Invalid training and testing percentage combination",
            ),
            (
                50,
                50,
                EXPANDING_WINDOW_STEPS,
                "Invalid training and testing percentage combination"
                f" given for {EXPANDING_WINDOW_STEPS} expanding steps",
            ),
            (50, 50, -1, "Invalid expanding steps"),
        ]

        for (
            start_train_p,
            test_p,
            expanding_steps,
            expected_msg,
        ) in forbidden_init_params:

            # Create backtester with forbidden initialization parameters
            with self.assertRaises(ValueError) as e:
                BackTesterExpandingWindow(
                    error_methods=ALL_ERRORS,
                    data=self.TSData,
                    params=self.model_params,
                    start_train_percentage=start_train_p,
                    end_train_percentage=PERCENTAGE,
                    test_percentage=test_p,
                    expanding_steps=expanding_steps,
                    model_class=self.model_class,
                    multi=False,
                )

            generated_msg = str(e.exception)
            self.assertEqual(expected_msg, generated_msg)

    def test_forbidden_train_test_splits(self) -> None:
        """
        Testing process consists of the following:
          1. Create a Backtest instance with allowed train test split values
          2. Set forbidden train test split values
          3. Ensure backtester raises an exception
        """

        # Create backtester for this given model class
        bt = BackTesterExpandingWindow(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            start_train_percentage=EXPANDING_WINDOW_START,
            end_train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            expanding_steps=EXPANDING_WINDOW_STEPS,
            model_class=self.model_class,
            multi=False,
        )

        # Create forbidden train test split values
        forbidden_train_test_splits = [
            (-10, 50, "Incorrect starting training size"),
            (60, 60, "Incorrect training and testing sizes"),
            (50, 50, "Incorrect training and testing sizes for multiple steps"),
        ]

        for train_p, test_p, expected_msg in forbidden_train_test_splits:
            # Set backtester with forbidden train test split values
            bt.start_train_percentage = train_p
            bt.test_percentage = test_p

            # Ensure backtester will raise an expection
            with self.assertRaises(ValueError) as e:
                bt._create_train_test_splits()

            generated_msg = str(e.exception)
            self.assertEqual(expected_msg, generated_msg)

    @classmethod
    def create_folds(
        cls,
        data: pd.DataFrame,
        start_train_size: float,
        end_train_size: float,
        test_size: float,
        num_folds: int,
    ) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
        """
        Ground truth fold creation
        """

        train_folds = []
        test_folds = []
        offsets = return_fold_offsets(int(start_train_size), int(end_train_size), num_folds)

        for offset in offsets:
            train_folds.append(data.iloc[: int(start_train_size + offset)])
            test_folds.append(data.iloc[int(start_train_size + offset) : int(start_train_size + offset + test_size)])
        return train_folds, test_folds


class RollingWindowBackTesterTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Setting up data
        DATA = load_air_passengers(return_ts=False)
        cls.TSData = load_air_passengers()

        cls.train_folds, cls.test_folds = cls.create_folds(
            data=DATA,
            train_size=ROLLING_WINDOW_TRAIN / 100.0 * len(DATA),
            test_size=(100 - PERCENTAGE) / 100.0 * len(DATA),
            num_folds=ROLLING_WINDOW_STEPS,
        )

        # Mock model results
        cls.model_class = mock.MagicMock()
        cls.model_class().predict.side_effect = [
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_18_90_FCST_DUMMY_DATA,
            PROPHET_36_108_FCST_DUMMY_DATA,
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_18_90_FCST_DUMMY_DATA,
            PROPHET_36_108_FCST_DUMMY_DATA,
        ]  # Once for backtester once for local model
        cls.model_params = mock.MagicMock()

    def test_error_values(self) -> None:
        """
        Testing process consists of the following:
          1. Backtest with mocked model
          2. Ensure backtester used correct train test splits
          3. Train (mocked) model locally and extract errors
          4. Comparing errors to backtester results
        """

        # Create and run backtester
        bt = BackTesterRollingWindow(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            train_percentage=ROLLING_WINDOW_TRAIN,
            test_percentage=(100 - PERCENTAGE),
            sliding_steps=ROLLING_WINDOW_STEPS,
            model_class=self.model_class,
            multi=False,
        )
        bt.run_backtest()

        # Test model initialization
        self.assertEqual(self.model_class.call_count, 4)
        self.model_class.assert_has_calls(
            [
                mock.call(data=TimeSeriesData(self.train_folds[0]), params=self.model_params),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(data=TimeSeriesData(self.train_folds[1]), params=self.model_params),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(data=TimeSeriesData(self.train_folds[2]), params=self.model_params),
            ]
        )

        # Test model predict
        # self.assertEqual(self.model_class().predict.call_count, 3)
        self.model_class().assert_has_calls(
            [
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
            ]
        )

        # Training local model and getting predictions
        true_errors = {}  # Dict to store errors
        for i in range(0, len(self.train_folds)):
            train_fold = self.train_folds[i]
            test_fold = self.test_folds[i]
            temp_model = self.model_class(data=TimeSeriesData(train_fold), params=self.model_params)
            temp_model.fit()

            # Getting model predictions
            temp_fcst = temp_model.predict(steps=TIMESTEPS, freq=FREQUENCY)

            # Using model predictions from local_model to calculate true errors
            pred = np.array(temp_fcst["fcst"])
            truth = np.array(test_fold["y"])
            train = np.array(train_fold["y"])
            compute_errors_list(train, pred, truth, true_errors)

        # Calculating errors
        for error_name, values in true_errors.items():
            true_errors[error_name] = statistics.mean(values)
        ground_truth_errors = (bt, true_errors)

        # # Test that local model errors equal backtester errors
        backtester, error_results = ground_truth_errors
        for error_name, value in error_results.items():
            self.assertEqual(
                round(value, FLOAT_ROUNDING_PARAM),
                round(backtester.errors[error_name], FLOAT_ROUNDING_PARAM),
            )

    def test_one_step_forecast(self) -> None:
        """
        Tests that if sliding steps is one, the folds returned are the same
        as BacktesterSimple.
        """

        # Calculate folds from Rolling Window Backtester
        rolling_backtester = BackTesterRollingWindow(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            sliding_steps=1,
            model_class=self.model_class,
            multi=False,
        )
        one_step_folds_rolling = rolling_backtester._create_train_test_splits()

        # Calculate folds from Simple Backtester
        simple_backtester = BackTesterSimple(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            train_percentage=PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            model_class=self.model_class,
        )
        folds_simple = simple_backtester._create_train_test_splits()

        # Test that folds are equivalent
        self.assertEqual(one_step_folds_rolling, folds_simple)

    @classmethod
    def create_folds(
        cls,
        data: pd.DataFrame,
        train_size: float,
        test_size: float,
        num_folds: int,
    ) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
        """
        Ground truth fold creation
        """

        offsets = return_fold_offsets(0, int(len(data) - train_size - test_size), num_folds)
        train_folds = []
        test_folds = []

        for offset in offsets:
            train_folds.append(data.iloc[offset : int(offset + train_size)])
            test_folds.append(data.iloc[int(offset + train_size) : int(offset + train_size + test_size)])
        return train_folds, test_folds


class FixedWindowBackTesterTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Setting up data
        DATA = load_air_passengers(return_ts=False)
        cls.TSData = load_air_passengers()

        # Creating folds
        cls.train_data = cls.TSData[: len(cls.TSData) - (TIMESTEPS * 2)]
        # pyre-fixme[6]: For 1st param expected `Optional[DataFrame]` but got
        #  `Union[DataFrame, Series]`.
        cls.test_data = TimeSeriesData(DATA.tail(TIMESTEPS))

    def prophet_predict_side_effect(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        if len(kwargs) > 1 and kwargs["steps"] == TIMESTEPS * 2 and kwargs["freq"] == FREQUENCY:
            return PROPHET_0_72_GAP_36_FCST_DUMMY_DATA
        else:
            return PROPHET_EMPTY_DUMMY_DATA

    def test_error_values(self) -> None:
        """
        Testing process consists of the following:
          1. Backtest with mocked model
          2. Ensure backtester used correct train test splits
          3. Train (mocked) model locally and extract errors
          4. Comparing errors to backtester results
        """

        # Mock model results
        model_class = mock.MagicMock()
        model_class().predict.side_effect = self.prophet_predict_side_effect
        model_params = mock.MagicMock()

        # Creating and running backtester for this given model class
        bt = BackTesterFixedWindow(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=model_params,
            train_percentage=FIXED_WINDOW_TRAIN_PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            window_percentage=FIXED_WINDOW_PERCENTAGE,
            model_class=model_class,
        )
        bt.run_backtest()

        # Testing that backtester initialized model with correct train/test split
        model_class.assert_called_with(data=self.train_data, params=model_params)

        # Training local model and getting predictions
        temp_model = model_class(data=self.train_data, params=model_params)
        temp_model.fit()
        temp_fcst = temp_model.predict(steps=TIMESTEPS * 2, freq=FREQUENCY)[TIMESTEPS:]

        # Using model predictions from local model to calculate true errors
        pred = np.array(temp_fcst["fcst"])
        truth = np.array(self.test_data.value)
        train = np.array(self.train_data.value)
        true_errors = compute_errors(train, pred, truth)
        ground_truth_errors = (bt, true_errors)

        # Test that local model errors equal backtester errors
        backtester, error_results = ground_truth_errors
        for error_name, value in error_results.items():
            self.assertEqual(
                round(value, FLOAT_ROUNDING_PARAM),
                round(backtester.errors[error_name], FLOAT_ROUNDING_PARAM),
            )

    def test_forbidden_initialization_parameters(self) -> None:
        """
        Testing process consists of the following:
          1. Create a Backtest instance with forbidden initial parameters
          2. Ensure backtester raises an exception
        """

        # Create backtester for this given model class
        # Mock model results
        model_class = mock.MagicMock()
        model_class().predict.side_effect = self.prophet_predict_side_effect
        model_params = mock.MagicMock()

        forbidden_init_params = [
            (-10, 50, FIXED_WINDOW_PERCENTAGE, "Invalid training percentage"),
            (110, 50, FIXED_WINDOW_PERCENTAGE, "Invalid training percentage"),
            (50, -10, FIXED_WINDOW_PERCENTAGE, "Invalid test percentage"),
            (50, 110, FIXED_WINDOW_PERCENTAGE, "Invalid test percentage"),
            (50, 50, -10, "Invalid window percentage"),
            (50, 50, 110, "Invalid window percentage"),
        ]

        for (train_p, test_p, window_p, expected_msg) in forbidden_init_params:

            # Create backtester with forbidden initialization parameters
            with self.assertRaises(ValueError) as e:
                BackTesterFixedWindow(
                    error_methods=ALL_ERRORS,
                    data=self.TSData,
                    params=model_params,
                    train_percentage=train_p,
                    test_percentage=test_p,
                    window_percentage=window_p,
                    model_class=model_class,
                )

            generated_msg = str(e.exception)
            self.assertEqual(expected_msg, generated_msg)

    def test_forbidden_train_test_splits(self) -> None:
        """
        Testing process consists of the following:
          1. Backtest with mocked model
          2. Set forbidden train test split values
          3. Ensure backtester raises an exception
        """

        # Create backtester for this given model class
        # Mock model results
        model_class = mock.MagicMock()
        model_class().predict.side_effect = self.prophet_predict_side_effect
        model_params = mock.MagicMock()

        # Creating and running backtester for this given model class
        bt = BackTesterFixedWindow(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=model_params,
            train_percentage=FIXED_WINDOW_TRAIN_PERCENTAGE,
            test_percentage=(100 - PERCENTAGE),
            window_percentage=FIXED_WINDOW_PERCENTAGE,
            model_class=model_class,
        )

        # Create forbidden train test split values
        forbidden_train_test_splits = [
            (-10, 50, 25, "Incorrect training size"),
            (50, -10, 25, "Incorrect testing size"),
            (50, 50, 25, "Incorrect training, testing, & window sizes"),
        ]

        for train_p, test_p, window_p, expected_msg in forbidden_train_test_splits:
            # Set backtester with forbidden train test split values
            bt.train_percentage = train_p
            bt.test_percentage = test_p
            bt.window_percentage - window_p

            # Ensure backtester will raise an expection
            with self.assertRaises(ValueError) as e:
                bt._create_train_test_splits()

            generated_msg = str(e.exception)
            self.assertEqual(expected_msg, generated_msg)


class CrossValidationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Setting up data
        cls.TSData = load_air_passengers()

        # Mock model and params
        cls.model_class = mock.MagicMock()
        cls.model_params = mock.MagicMock()

    def test_error_values_expanding(self) -> None:
        """
        Tests Expanding Window Cross Validation

        Testing process consists of the following:
          1. Creates folds from the data.
          2. Trains local model on folds.
          3. Calculates local model predictions and errors.
          4. Runs CV class on the data.
          5. Tests CV initialization and model prediction
          6. Compares CV errors to local model errors.
        """

        # Create folds
        DATA = self.TSData.to_dataframe()
        expanding_train_folds, expanding_test_folds = self.create_folds(
            data=DATA,
            train_size=EXPANDING_WINDOW_START / 100 * len(DATA),
            num_folds=CV_NUM_FOLDS,
            test_size=(100 - PERCENTAGE) / 100.0 * len(DATA),
            expanding=True,
        )

        # Create data structures to store CV results
        expanding_cv_results = {}
        true_errors = {}

        # Mock model results
        self.model_class().predict.side_effect = [
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_0_90_FCST_DUMMY_DATA,
            PROPHET_0_108_FCST_DUMMY_DATA,
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_0_90_FCST_DUMMY_DATA,
            PROPHET_0_108_FCST_DUMMY_DATA,
        ]  # Once for ground truth once for CV

        # Iterate through folds and train model to produce ground truth
        for i in range(0, len(expanding_train_folds)):
            train_fold = expanding_train_folds[i]
            test_fold = expanding_test_folds[i]
            temp_model = self.model_class(
                data=TimeSeriesData(train_fold),
                params=self.model_params,
            )
            temp_model.fit()
            # Getting model predictions
            temp_fcst = temp_model.predict(steps=TIMESTEPS, freq=FREQUENCY)

            # Using model predictions from temp_model to calculate true errors
            pred = np.array(temp_fcst["fcst"])
            truth = np.array(test_fold["y"])
            train = np.array(train_fold["y"])
            compute_errors_list(train, pred, truth, true_errors)

        # Calculate average error across folds
        for error_name, values in true_errors.items():
            true_errors[error_name] = statistics.mean(values)

        # Creating and running CV Object for this given model class
        temp_cv = CrossValidation(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            train_percentage=EXPANDING_WINDOW_START,
            test_percentage=(100 - PERCENTAGE),
            num_folds=CV_NUM_FOLDS,
            model_class=self.model_class,
            multi=False,
        )
        temp_cv.run_cv()
        expanding_cv_results = (temp_cv, true_errors)

        # Test CV initialization
        self.model_class.assert_has_calls(
            [
                mock.call(
                    data=TimeSeriesData(expanding_train_folds[0]),
                    params=self.model_params,
                ),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(
                    data=TimeSeriesData(expanding_train_folds[1]),
                    params=self.model_params,
                ),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(
                    data=TimeSeriesData(expanding_train_folds[2]),
                    params=self.model_params,
                ),
            ]
        )

        # Testing CV predict
        # self.assertEqual(self.model_class().predict.call_count, 6)
        self.model_class().assert_has_calls(
            [
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
            ]
        )

        # Test that CV errors equal model prediction errors
        cv, error_results = expanding_cv_results
        for error_name, value in error_results.items():
            self.assertEqual(
                round(value, FLOAT_ROUNDING_PARAM),
                round(cv.errors[error_name], FLOAT_ROUNDING_PARAM),
            )

    # Tests backtester error values against "true" error values
    def test_error_values_rolling(self) -> None:
        """
        Tests Rolling Window Cross Validation

        Testing process consists of the following:
          1. Creates folds from the data.
          2. Trains local model on folds.
          3. Calculates local model predictions and errors.
          4. Runs CV class on the data.
          5. Tests CV initialization and model prediction
          6. Compares CV errors to local model errors.
        """

        # Create folds
        DATA = self.TSData.to_dataframe()
        rolling_train_folds, rolling_test_folds = self.create_folds(
            data=DATA,
            train_size=EXPANDING_WINDOW_START / 100 * len(DATA),
            num_folds=CV_NUM_FOLDS,
            test_size=(100 - PERCENTAGE) / 100.0 * len(DATA),
            expanding=False,
        )

        # Create data structures to store CV results
        rolling_cv_results = {}
        true_errors = {}

        # Mock model results
        self.model_class().predict.side_effect = [
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_18_90_FCST_DUMMY_DATA,
            PROPHET_36_108_FCST_DUMMY_DATA,
            PROPHET_0_72_FCST_DUMMY_DATA,
            PROPHET_18_90_FCST_DUMMY_DATA,
            PROPHET_36_108_FCST_DUMMY_DATA,
        ]  # Once for ground truth once for CV

        # Iterate through folds and train model to produce ground truth
        for i in range(0, len(rolling_train_folds)):
            train_fold = rolling_train_folds[i]
            test_fold = rolling_test_folds[i]
            temp_model = self.model_class(
                data=TimeSeriesData(train_fold),
                params=self.model_params,
            )
            temp_model.fit()
            # Getting model predictions
            temp_fcst = temp_model.predict(steps=TIMESTEPS, freq=FREQUENCY)

            # Using model predictions from temp_model to calculate true errors
            pred = np.array(temp_fcst["fcst"])
            truth = np.array(test_fold["y"])
            train = np.array(train_fold["y"])
            compute_errors_list(train, pred, truth, true_errors)

        # Calculate average error across folds
        for error_name, values in true_errors.items():
            true_errors[error_name] = statistics.mean(values)

        # Creating and running CV Object for this given model class
        temp_cv = CrossValidation(
            error_methods=ALL_ERRORS,
            data=self.TSData,
            params=self.model_params,
            train_percentage=ROLLING_WINDOW_TRAIN,
            test_percentage=(100 - PERCENTAGE),
            num_folds=CV_NUM_FOLDS,
            model_class=self.model_class,
            constant_train_size=True,
            multi=False,
        )
        temp_cv.run_cv()
        rolling_cv_results = (temp_cv, true_errors)

        # Test CV initialization
        self.model_class.assert_has_calls(
            [
                mock.call(
                    data=TimeSeriesData(rolling_train_folds[0]),
                    params=self.model_params,
                ),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(
                    data=TimeSeriesData(rolling_train_folds[1]),
                    params=self.model_params,
                ),
                mock.call().fit(),
                mock.call().predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call(
                    data=TimeSeriesData(rolling_train_folds[2]),
                    params=self.model_params,
                ),
            ]
        )

        # Testing CV predict
        self.model_class().assert_has_calls(
            [
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
                mock.call.fit(),
                mock.call.predict(steps=TIMESTEPS, freq=FREQUENCY),
            ]
        )

        # Test that CV errors equal model prediction errors
        cv, error_results = rolling_cv_results
        for error_name, value in error_results.items():
            self.assertEqual(
                round(value, FLOAT_ROUNDING_PARAM),
                round(cv.errors[error_name], FLOAT_ROUNDING_PARAM),
            )

    def test_forbidden_initialization_parameters(self) -> None:
        """
        Testing process consists of the following:
          1. Create a cross validation instance with forbidden initial parameters
          2. Ensure cross validation raises an exception
        """

        forbidden_init_params = [
            (-10, 50, CV_NUM_FOLDS, "Invalid training percentage"),
            (110, 50, CV_NUM_FOLDS, "Invalid training percentage"),
            (50, -10, CV_NUM_FOLDS, "Invalid test percentage"),
            (50, 110, CV_NUM_FOLDS, "Invalid test percentage"),
            (50, 50, -10, "Invalid number of folds"),
        ]

        for (train_p, test_p, num_folds, expected_msg) in forbidden_init_params:

            # Create cross validation object with forbidden initialization parameters
            with self.assertRaises(ValueError) as e:
                CrossValidation(
                    error_methods=ALL_ERRORS,
                    data=self.TSData,
                    params=self.model_params,
                    train_percentage=train_p,
                    test_percentage=test_p,
                    num_folds=num_folds,
                    model_class=self.model_class,
                    multi=False,
                )

            generated_msg = str(e.exception)
            self.assertEqual(expected_msg, generated_msg)

    def create_folds(
        self,
        data: pd.DataFrame,
        train_size: float,
        num_folds: int,
        test_size: float,
        expanding: bool,
    ) -> Tuple[List[pd.DataFrame], List[pd.DataFrame]]:
        train_folds = []
        test_folds = []
        """
        Ground truth fold creation
        """

        if expanding:
            end_train_size = len(data) - test_size
            window_size = (end_train_size - train_size) / (num_folds - 1)
            for i in range(0, num_folds):
                end_train_range = int((i * window_size) + train_size)
                train_folds.append(data.iloc[:end_train_range])
                test_folds.append(data.iloc[end_train_range : end_train_range + int(test_size)])
            return train_folds, test_folds

        fold_size = (len(data) - train_size - test_size) / (num_folds - 1)

        for i in range(0, num_folds):
            offset = int(i * fold_size)
            train_folds.append(data.iloc[offset : int(offset + train_size)])
            test_folds.append(data.iloc[int(offset + train_size) : int(offset + train_size + test_size)])
        return train_folds, test_folds


class KatsSimpleBacktesterTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Setting up data
        cls.ts = TimeSeriesData(time=pd.date_range("2022-05-06", periods=20), value=pd.Series(np.arange(20)))

    def side_effect(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        n = kwargs["steps"]
        return pd.DataFrame({"y": np.arange(n), "fcst": np.arange(n)})

    def test_backtester(self) -> None:
        # mock forecasting model
        model_class = mock.MagicMock()
        model_class().predict.side_effect = self.side_effect
        model_params = mock.MagicMock()
        # initiate data partition
        dp = SimpleDataPartition(train_frac=0.9)
        # run backtester
        bt = KatsSimpleBacktester(
            datapartition=dp,
            # pyre-fixme Incompatible parameter type [6]: In call `KatsSimpleBacktester.__init__`, for 2nd parameter `scorer` expected `Union[List[Union[Metric, MultiOutputMetric, WeightedMetric, str]], Metric, MultiOutputMetric, WeightedMetric, Scorer, str]` but got `List[str]`.
            scorer=ALL_ERRORS,
            model_params=model_params,
            model_class=model_class,
        )
        bt.run_backtester(self.ts)
        bt_res = cast(BacktesterResult, bt.get_errors())
        # get errors directly via model
        errors = compute_errors(np.arange(18), np.arange(2), np.arange(18, 20))
        self.assertEqual(bt_res.fold_errors, [errors])
        self.assertEqual(bt_res.summary_errors, errors)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""ARIMA (Auto Regressive Integrated Moving Average) for time series data.

ARIMA is a classical statistical model with three main components from its name:
- AR, Auto Regressive, means the variable of interest (time series) is regressed
  on its own lagged values
- I, Integrated, means data values have been replaced with the difference between
  their values and the previous value
- MA, Moving Average, means the regression error is a linear combination of error
  terms whose values occurred contemporaneously and at various times in the past

We use the implementation in statsmodels
<https://www.statsmodels.org/stable/index.html> and re-write the API to fit
Kats development style.
"""

import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
from statsmodels.tsa.arima_model import ARIMA
from statsmodels.tsa.arima_model import ARIMAResults

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model
from kats.utils.parameter_tuning_utils import get_default_arima_parameter_search_space


class ARIMAParams(Params):
    """Parameter class for ARIMA model

    This is the parameter class for ARIMA model, it contains all necessary
    parameters from the following ARIMA implementation:
    https://www.statsmodels.org/stable/generated/statsmodels.tsa.arima_model.ARIMA.html

    Attributes:
        p: An integer for trend autoregressive (AR) order
        d: An integer for trend difference order
        q: An integer for trend moving average (MA) order
        exog: Optional; An array of exogenous regressors
        dates: Optional; pandas-compatible datetime object
        freq: Optional; frequency of a given time series
    """

    p: int
    d: int
    q: int
    exog: Optional[np.ndarray] = None
    dates: Optional[pd.DatetimeIndex] = None
    freq: Optional[str] = None

    def __init__(self, p: int, d: int, q: int, **kwargs: Any) -> None:
        super().__init__()
        self.p = p
        self.d = d
        self.q = q
        self.exog = kwargs.get("exog", None)
        self.dates = kwargs.get("dates", None)
        self.freq = kwargs.get("freq", None)
        logging.debug("Initialized ARIMAParams with parameters. " f"p:{p}, d:{d}, q:{q}, kwargs:{kwargs}")

    def validate_params(self) -> None:
        logging.info("Method validate_params() is not implemented.")
        pass


class ARIMAModel(Model[ARIMAParams]):
    """Model class for ARIMA model

    Attributes:
        data: :class:`kats.consts.TimeSeriesData`, the input historical time
            series data from TimeSeriesData
        params: The ARIMA model parameters from ARIMAParams
    """

    exog: Optional[int] = None
    alpha: float = 0.05
    freq: Optional[str] = None
    model: Optional[ARIMAResults] = None
    include_history: bool = False
    fcst_df: Optional[pd.DataFrame] = None
    y_fcst: Optional[np.ndarray] = None
    y_fcst_lower: Optional[np.ndarray] = None
    y_fcst_upper: Optional[np.ndarray] = None
    start_params: Optional[np.ndarray] = None
    transparams: bool = False
    method: Optional[str] = None
    trend: Optional[str] = None
    solver: Optional[str] = None
    maxiter: Optional[int] = None
    full_output: bool = False
    disp: Optional[int] = None
    callback: Optional[Callable[[np.ndarray], None]] = None
    start_ar_lags: Optional[int] = None
    dates: Optional[pd.DatetimeIndex] = None

    def __init__(self, data: TimeSeriesData, params: ARIMAParams) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but got " f"{type(self.data.value)}."
            logging.error(msg)
            raise ValueError(msg)

    def fit(
        self,
        start_params: Optional[np.ndarray] = None,
        transparams: bool = True,
        method: str = "css-mle",
        trend: str = "c",
        solver: str = "lbfgs",
        maxiter: int = 500,
        full_output: bool = True,
        disp: int = 5,
        callback: Optional[Callable[[np.ndarray], None]] = None,
        start_ar_lags: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """Fit ARIMA model with given parameters

        For more details on each parameter please refer to the following doc:
        https://www.statsmodels.org/stable/generated/statsmodels.tsa.arima_model.ARIMA.fit.html#statsmodels.tsa.arima_model.ARIMA.fit

        Args:
            start_params: Optional; An array_like object for the initial guess
                of the solution for the loglikelihood maximization
            transparams: Optional; A boolean to specify whether or not to
                transform the parameters to ensure stationarity.
            method: A string that specifies the loglikelihood to maximize. Can
                be 'css-mle', 'mle' and 'css'.
            trend: A string that specifies the whether to include a constant in
                the trend or not. Can be 'c' and 'nc'.
            solver: Optional; A string that specifies specifies the solver to be
                used. Can be 'bfgs', 'newton', 'cg', 'ncg' and 'powell'.
            maxiter: Optional; A integer for the maximum number of function
                iterations.
            tol: Optional; The convergence tolerance for the fitting.
            full_output: Optional; A boolean to specify whether to show all
                output from the solver in the results.
            disp: Optional; A integer to control the frequency of the output
                during the iterations.
            callback: Optional; A callable object to be called after each iteration.
            start_ar_lags Optional; An integer to specify the AR lag parameter
                to fit the start_params.

        Returns:
            None
        """

        logging.debug("Call fit() method")
        self.start_params = start_params
        self.transparams = transparams
        self.method = method
        self.trend = trend
        self.solver = solver
        self.maxiter = maxiter
        self.full_output = full_output
        self.disp = disp
        self.callback = callback
        self.start_ar_lags = start_ar_lags

        arima = ARIMA(
            # pyre-fixme[16]: `Optional` has no attribute `value`.
            self.data.value,
            order=(self.params.p, self.params.d, self.params.q),
            exog=self.params.exog,
            # pyre-fixme[16]: `Optional` has no attribute `time`.
            dates=self.data.time,
            freq=self.params.freq,
        )
        logging.info("Created arima model.")
        self.model = arima.fit(
            start_params=self.start_params,
            transparams=self.transparams,
            method=self.method,
            trend=self.trend,
            solver=self.solver,
            maxiter=self.maxiter,
            full_output=self.full_output,
            disp=self.disp,
            callback=self.callback,
            start_ar_lags=self.start_ar_lags,
        )
        logging.info("Fitted arima.")

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, include_history: bool = False, **kwargs: Any) -> pd.DataFrame:
        """Predict with fitted ARIMA model

        Args:
            steps: An integer for forecast steps
            include_history: Optional; A boolean to specify whether to include
                historical data.

        Returns:
            A pd.DataFrame that contains the forecast and confidence interval
        """
        model = self.model
        if model is None:
            raise ValueError("Call fit() before predict().")

        logging.debug(f"predict(steps:{steps}, kwargs:{kwargs})")
        self.include_history = include_history
        self.exog = kwargs.get("exog", None)
        self.alpha = kwargs.get("alpha", 0.05)
        # pyre-fixme[16]: `Optional` has no attribute `time`.
        self.freq = kwargs.get("freq", pd.infer_freq(self.data.time))
        fcst = model.forecast(steps, exog=self.exog, alpha=self.alpha)
        logging.info("Generated forecast data from arima model.")
        logging.debug(f"Forecast data: {fcst}")

        self.y_fcst = fcst[0]
        lower, upper = fcst[2].transpose()
        self.y_fcst_lower = lower
        self.y_fcst_upper = upper

        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)

        self.dates = dates[dates != last_date]  # Return correct number of periods

        self.fcst_df = fcst_df = pd.DataFrame(
            {
                "time": self.dates,
                "fcst": self.y_fcst,
                "fcst_lower": lower,
                "fcst_upper": upper,
            },
            copy=False,
        )
        if self.include_history:
            try:
                hist_fcst = (
                    # pyre-fixme[6]: For 1st param expected `Sized` but got
                    #  `Optional[TimeSeriesData]`.
                    model.predict(self.params.d, len(self.data))
                    .reset_index()
                    .rename(columns={"index": "time", 0: "fcst"})
                )
                self.fcst_df = fcst_df = pd.concat([hist_fcst, fcst_df], copy=False)
            except Exception as e:
                msg = "Fail to generate in-sample forecasts for historical data " f"with error message {e}."
                logging.error(msg)
                raise ValueError(msg)
        logging.debug(f"Return forecast data: {fcst_df}")
        return fcst_df

    def __str__(self) -> str:
        return "ARIMA"

    @staticmethod
    # pyre-fixme[15]: `get_parameter_search_space` overrides method defined in
    #  `Model` inconsistently.
    def get_parameter_search_space() -> List[Dict[str, Any]]:
        """Get default ARIMA parameter search space.

        Args:
            None

        Returns:
            A dictionary with the default ARIMA parameter search space
        """

        return get_default_arima_parameter_search_space()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""This module contains the class TemporalHierarchicalModel class.
"""

import logging
from math import gcd
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

import numpy as np
import pandas as pd
from sklearn.covariance import MinCovDet

from kats.consts import TimeSeriesData
from kats.metrics import metrics
from kats.models import arima
from kats.models import holtwinters
from kats.models import linear_model
from kats.models import quadratic_model
from kats.models import sarima
from kats.models import theta
from kats.models.model import Model
from kats.models.reconciliation.base_models import BaseTHModel
from kats.models.reconciliation.base_models import GetAggregateTS

# pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
BASE_MODELS: Dict[str, Type[Model]] = {
    "arima": arima.ARIMAModel,
    "holtwinters": holtwinters.HoltWintersModel,
    "sarima": sarima.SARIMAModel,
    "linear": linear_model.LinearModel,
    "quadratic": quadratic_model.QuadraticModel,
    "theta": theta.ThetaModel,
}


def _log_error(msg: str) -> ValueError:
    logging.error(msg)
    return ValueError(msg)


class TemporalHierarchicalModel:
    """Temporal hierarchical model class.

    This framework combines the base models of different temporal aggregation
    levels to generate reconciled forecasts.
    This class provides fit, get_S, get_W, predict and median_validation.

    Attributes:
        data: A TimeSeriesData object storing the time series data for level 1
            (i.e., the most disaggregate level).
        baseModels: A list BaseTHModel objects representing the base models for
            different levels.
    """

    # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
    models: Optional[Dict[str, Model]] = None
    residuals: Optional[Dict[int, np.ndarray]] = None
    res_matrix: Optional[np.ndarray] = None

    def __init__(self, data: TimeSeriesData, baseModels: List[BaseTHModel]) -> None:

        if not data.is_univariate():
            msg = f"Only univariate time series supported, but got {type(data.value)}."
            raise _log_error(msg)

        self.data = data
        for basemodel in baseModels:
            if not isinstance(basemodel, BaseTHModel):
                msg = "Base model should be a BaseTHModel object but is " f"{type(basemodel)}."
                raise _log_error(msg)

        levels = [bm.level for bm in baseModels]

        if 1 not in levels:
            raise _log_error("Model of level 1 is missing.")

        if len(levels) != len(set(levels)):
            raise _log_error("One level cannot receive multiple models.")

        # pyre-fixme[4]: Attribute must be annotated.
        self.levels = sorted(levels, reverse=True)

        m = self._get_m(levels)
        # pyre-fixme[4]: Attribute must be annotated.
        self.m = m
        # pyre-fixme[4]: Attribute must be annotated.
        self.freq = {k: int(m / k) for k in self.levels}
        self.baseModels = baseModels
        # pyre-fixme[4]: Attribute must be annotated.
        self.info_fcsts = {}
        # pyre-fixme[4]: Attribute must be annotated.
        self.info_residuals = {}

    def _get_m(self, ks: List[int]) -> int:
        """Calculate m.

        m is the minimum common multiple of all levels.

        Args:
            ks: the list of integers representing all the levels.

        Returns:
            An integer representing the minimum common multiple.
        """

        base = 1
        for c in ks:
            base = base * c // gcd(base, c)
        return base

    def fit(self) -> None:
        """Fit all base models.

        If base model only has residuals and forecasts, store the information.
        """

        levels = self.levels
        TSs = GetAggregateTS(self.data).aggregate(levels)
        models = {}
        residuals = {}
        fcsts = {}
        for bm in self.baseModels:
            model_name = bm.model_name
            if model_name is None:  # only residuals and fcsts are provided
                models[bm.level] = None
                residuals[bm.level] = bm.residuals
                fcsts[bm.level] = bm.fcsts
            else:
                m = BASE_MODELS[model_name](
                    data=TSs[bm.level],
                    params=bm.model_params,
                )
                m.fit()
                models[bm.level] = m
        self.models = models
        self.info_fcsts = fcsts
        self.info_residuals = residuals

    def get_S(self) -> np.ndarray:
        """Calculate S matrix.

        Returns:
            A np.array representing the S matrix.
        """

        ans = []
        levels = self.levels
        m = self.m
        for k in levels:
            for i in range(self.freq[k]):
                tem = np.zeros(m)
                tem[(i * k) : (i * k + k)] = 1.0
                ans.append(tem)
        return np.row_stack(ans)

    def _aggregate_data(self, data: np.ndarray, k: int) -> np.ndarray:
        """Aggregate data according to level k."""

        if k == 1:
            return data
        n = len(data)
        h = n // k
        return (data[: int(h * k)]).reshape(-1, k).sum(axis=1)

    # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
    def _get_residuals(self, model: Model) -> np.ndarray:
        """Calculate residuals of each base model.

        Args:
            model: a callable model object representing the trained base model.

        Returns:
            A np.ndarray of residuals.
        """
        try:
            # pyre-fixme[16]: `Model` has no attribute `model`.
            return model.model.resid.values
        except Exception:
            fcst = model.predict(steps=1, freq="D", include_history=True)
            # pyre-fixme[16]: `None` has no attribute `merge`.
            # pyre-fixme[16]: `Optional` has no attribute `to_dataframe`.
            merge = fcst.merge(model.data.to_dataframe(), on="time")
            for col in merge.columns:
                if col != "time" and ("fcst" not in col):
                    return merge[col].values - merge["fcst"].values
            raise ValueError("Couldn't find residual or forecast values in model")

    def _get_all_residuals(self) -> Dict[int, np.ndarray]:
        """
        Calculate residuals for all base models.

        Returns:
            Dictionary for residuals, whose key is level and value is residual
            array.
        """
        residuals = self.residuals
        # if residuals have not been calculated yet
        if residuals is None:
            levels = self.levels
            models = self.models
            residuals = {}
            for k in levels:
                # assert models is not None
                # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
                if models[k] is not None:
                    try:
                        vals = self._get_residuals(models[k])
                    except Exception as e:
                        msg = f"Failed to get residuals for level {k} with error " f"message {e}."
                        raise _log_error(msg)

                    residuals[k] = vals
                else:
                    residuals[k] = self.info_residuals[k]
            self.residuals = residuals
        return residuals

    def _get_residual_matrix(self) -> np.ndarray:
        """
        Reshape residuals into matrix format.

        Returns:
            Residual matrix.
        """
        res_matrix = self.res_matrix
        if res_matrix is None:
            residuals = self._get_all_residuals()
            ks = self.levels
            freq = self.freq
            h = np.min([len(residuals[k]) // freq[k] for k in ks])
            res_matrix = []
            for k in ks:
                n = h * freq[k]
                res_matrix.append(residuals[k][-n:].reshape(h, -1).T)
            res_matrix = np.row_stack(res_matrix)
            self.res_matrix = res_matrix
        return res_matrix

    def get_W(self, method: str = "struc", eps: float = 1e-5) -> np.ndarray:
        """
        Calculate W matrix.

        Args:
            method: Reconciliation method for temporal hierarchical model. Valid
                methods include 'struc', 'svar', 'hvar', 'mint_sample', and
                'mint_shrink'.
            eps: Epsilons added to W for numerical stability.

        Returns:
            W matrix. (If W is a diagnoal matrix, only returns its diagnoal elements).
        """
        levels = self.levels
        freq = self.freq
        if method == "struc":
            ans = []
            for k in levels:
                ans.extend([k] * freq[k])
            return np.array(ans)

        elif method == "svar":
            residuals = self._get_all_residuals()
            ans = []
            for k in levels:
                ans.extend([np.nanmean(np.square(residuals[k]))] * freq[k])
            return np.array(ans) + eps

        elif method == "hvar":
            res_matrix = self._get_residual_matrix()
            return np.nanvar(res_matrix, axis=1) + eps

        elif method == "mint_shrink":
            cov = np.cov(self._get_residual_matrix())
            # get correlation matrix
            sqrt = np.sqrt(np.diag(cov))
            cor = ((cov / sqrt).T) / sqrt  # due to symmetry, no need to transpose the matrix again.
            mask = ~np.eye(cor.shape[0], dtype=bool)
            cor = cor[mask]
            lam = np.var(cor) / np.sum(cor**2)
            lam = np.max([0, lam])
            cov = np.diag(np.diag(cov)) * lam + (1.0 - lam) * cov
            cov += np.eye(len(cov)) * eps
            return cov

        elif method == "mint_sample":
            cov = np.cov(self._get_residual_matrix())
            cov += np.eye(len(cov)) * eps
            return cov

        else:
            raise _log_error(f"{method} is invalid for get_W() method.")

    # pyre-fixme[2]: Parameter must be annotated.
    def _predict_origin(self, steps: int, method="struc") -> Dict[int, np.ndarray]:
        """
        Generate original forecasts from each base model (without time index).

        Args:
            steps: Number of forecasts for level 1.
            methd: Reconciliation method.

        Returns:
            Dictionary of forecasts of each level, whose key is level and value
            is forecast array.
        """
        m = self.m
        levels = self.levels
        freq = self.freq
        h = int(np.ceil(steps / m))
        hf = steps // m
        orig_fcst = {}
        models = self.models
        # generate forecasts for each level
        for k in levels:
            num = int(freq[k] * h)
            # assert models is not None
            # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
            if models[k] is not None:
                orig_fcst[k] = models[k].predict(steps=num, freq="D")["fcst"].values
            else:
                fcst_num = len(self.info_fcsts[k])
                if fcst_num < num:
                    if fcst_num >= hf * freq[k]:
                        # since the final output only needs hf*freq[k] forecasts
                        # for level k, we pad the forecast array to desirable
                        # length. (note that the padding values would be ignored
                        # in the final output.)
                        orig_fcst[k] = np.concatenate(
                            [
                                self.info_fcsts[k],
                                [self.info_fcsts[k][-1]] * (num - fcst_num),
                            ]
                        )

                    elif method == "bu" and k != 1:
                        # for 'bu' only level 1 is needed.
                        orig_fcst[k] = self.info_fcsts[k]
                    else:
                        msg = (
                            f"{hf*freq[k]} steps of forecasts for level {k} are"
                            f" needed, but only receive {fcst_num} steps (and "
                            "forecast model is None)."
                        )
                        raise _log_error(msg)
                else:
                    orig_fcst[k] = self.info_fcsts[k][:num]
        return orig_fcst

    def _predict(
        self,
        steps: int,
        # pyre-fixme[2]: Parameter must be annotated.
        method="struc",
        origin_fcst: bool = False,
        fcst_levels: Optional[List[int]] = None,
    ) -> Dict[str, Dict[int, np.ndarray]]:
        """Generate forecasts for each level (without time index).

        Args:
            steps: Number of forecasts for level 1.
            methd: Reconciliation method.
            origin_fcst: Whether to return the forecasts of base models.
            fcst_levels: Levels that one wants to generate forecasts for.
                If None, then all forecasts for all levels of the base models
                are generated.

        Returns:
            Dictionary of forecasts, whose key is level and value is
            forecast array.
        """
        if self.models is None:
            raise _log_error("Please fit base models via .fit() first.")

        m = self.m
        levels = self.levels
        freq = self.freq
        h = int(np.ceil(steps / m))
        if fcst_levels is None:
            fcst_levels = list(levels)
        fcst = {}
        orig_fcst = self._predict_origin(steps, method)

        if method == "bu":
            # bottom_up method
            yhat = orig_fcst[1]
        elif method == "median":
            # median method
            tem = []
            for k in levels:
                tem.append(np.repeat(orig_fcst[k] / k, k))
            tem = np.row_stack(tem)
            yhat = np.median(tem, axis=0)
        elif method in {"struc", "svar", "hvar", "mint_shrink", "mint_sample"}:
            # transform fcsts into matrix
            yh = []
            for k in levels:
                yh.append(orig_fcst[k].reshape(h, -1).T)
            yh = np.row_stack(yh)
            S = self.get_S()
            W = self.get_W(method)
            # when W is a vector, i.e., a simpler represent for a diagnoal matrix
            if len(W.shape) == 1:
                T = (S.T) / W
            else:
                T = np.linalg.solve(W, S).T
            yhat = np.dot(S, np.linalg.solve(T.dot(S), T)).dot(yh)
            # extract forecasts for level 1
            yhat = (yhat[(-freq[1]) :, :].T).flatten()[:steps]
        else:
            raise _log_error(f"Reconciliation method {method} is invalid.")
        # aggregate fcsts
        for k in fcst_levels:
            fcst[k] = self._aggregate_data(yhat, k)[: (steps // k)]
        ans = {"fcst": fcst}
        if origin_fcst:
            for elm in orig_fcst:
                orig_fcst[elm] = orig_fcst[elm][: (steps // elm)]
            ans["origin_fcst"] = orig_fcst
        return ans

    def predict(
        self,
        steps: int,
        # pyre-fixme[2]: Parameter must be annotated.
        method="struc",
        freq: Optional[str] = None,
        origin_fcst: bool = False,
        fcst_levels: Optional[List[int]] = None,
        last_timestamp: Optional[pd.Timestamp] = None,
    ) -> Dict[str, Dict[int, pd.DataFrame]]:
        """Generate reconciled forecasts (with time index).

        Args:
            steps: The number of forecasts needed for level 1.
            methd: The name of the reconciliation method. Can be 'bu'
                (bottom-up), 'median', 'struc' (structure-variance), 'svar',
                'hvar', 'mint_shrink' or 'mint_sample'.
            freq: The frequency of the time series at level 1. If None, then we
                infer the frequency via ts.infer_freq_robust().
            origin_fcst: Whether or not to return the forecasts of base models.
            fcst_levels: The levels to generate forecasts for. Default is None,
                which generates forecasts for all the levels of the base models.

        Returns:
            A dictionary of forecasts, whose key is the level and the
            corresponding value is a np.array storing the forecasts.
        """

        if freq is None:
            freq = self.data.infer_freq_robust()
        last_timestamp = self.data.time.max()
        fcsts = self._predict(steps, method=method, origin_fcst=origin_fcst, fcst_levels=fcst_levels)
        ans = {}
        for elm in fcsts:
            tmp = {}
            for k in fcsts[elm]:
                fcst_num = len(fcsts[elm][k])
                time = pd.date_range(
                    last_timestamp + freq * k,
                    last_timestamp + freq * k * fcst_num,
                    periods=fcst_num,
                )
                tmp[k] = pd.DataFrame({"time": time, "fcst": fcsts[elm][k]}, copy=False)
            ans[elm] = tmp
        return ans

    def median_validation(
        self,
        # pyre-fixme[2]: Parameter must be annotated.
        steps,
        dist_metric: str = "mae",
        threshold: float = 5.0,
    ) -> List[int]:
        """Filtering out bad fcsts based on median forecasts.

        This function detects the levels whose forecasts are greatly deviate
        from median forecasts, which is a strong indication of bad forecasts.

        Args:
            steps: The number of forecasts needed for level 1 for validation.
            dist_metric: The distance metric used to measure the distance between
                the base forecasts and the median forecasts.
            threshold: The threshold for deviance. The forecast whose distance
                from the median forecast is greater than threshold*std is taken
                as bad forecasts. Default is 3.

        Returns:
            A list of integers representing the levels whose forecasts are bad.
        """

        diffs = {}
        ks = self.levels
        try:
            func = metrics.core_metric(dist_metric)
        except ValueError:
            raise _log_error(f"Invalid dist_metric {dist_metric}")
        median_fcst = self._predict(steps, method="median", origin_fcst=True)
        for k in ks:
            diffs[k] = func(median_fcst["fcst"][k], median_fcst["origin_fcst"][k])
            if dist_metric == "mae":
                diffs[k] /= k
        vals = np.array(list(diffs.values()))
        try:
            cov = MinCovDet().fit(vals[:, None])
            lqr = np.sqrt(cov.covariance_.flatten()[0])
        except Exception:
            low, up = np.percentile(vals, [25, 75])
            lqr = up - low
        up = np.median(vals) + lqr * threshold
        return [k for k in diffs if diffs[k] >= up]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""STLF forecasting model

This model starts from decomposing the time series data with STL decomposition
then it fits individual foreasting model on the de-seasonalized components
it re-seasonalizes the forecasted results with seasonal data to produce the final
forecasting results.
"""

from __future__ import absolute_import
from __future__ import annotations
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import math
import operator
import operator as _operator
from copy import copy
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from typing import cast

import numpy as np
import pandas as pd

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models import linear_model
from kats.models import quadratic_model
from kats.models import simple_heuristic_model
from kats.models import theta
from kats.models.linear_model import LinearModel
from kats.models.linear_model import LinearModelParams
from kats.models.model import Model
from kats.models.quadratic_model import QuadraticModel
from kats.models.quadratic_model import QuadraticModelParams
from kats.models.simple_heuristic_model import SimpleHeuristicModel
from kats.models.simple_heuristic_model import SimpleHeuristicModelParams
from kats.models.theta import ThetaModel
from kats.models.theta import ThetaParams
from kats.utils.decomposition import TimeSeriesDecomposition
from kats.utils.parameter_tuning_utils import get_default_stlf_parameter_search_space

MODELS = ["linear", "quadratic", "theta", "simple"]


class STLFParams(Params):
    """Parameter class for STLF model

    This is the parameter class for STLF model, stands for STL-decomposition based
    forecasting model.

    Attributes:
        method: str, the forecasting model to fit on the de-seasonalized component
            it currently supports linear, quadratic, and theta method.
        m: int, the length of one seasonal cycle, same as period in statsmodel STL
        method_params: Optional[Params], the parameters for the method
        decomposition: str, `additive` or `multiplicative` decomposition. Default is `multiplicative` because of legacy
        seasonal: int, Length of the seasonal smoother. Must be an odd integer, and should normally be >= 7 (default).
        trend: Optional[int], Length of the trend smoother. Must be an odd integer.
              If not provided uses the smallest odd integer greater than 1.5 * period / (1 - 1.5 / seasonal),
              following the suggestion in the original implementation.
        low_pass: Optional[int], Length of the low-pass filter. Must be an odd integer >=3. If not provided, uses the smallest odd integer > period.
        seasonal_deg: int, Degree of seasonal LOESS. 0 (constant) or 1 (constant and trend).
        trend_deg: int, Degree of trend LOESS. 0 (constant) or 1 (constant and trend).
        low_pass_deg: int, Degree of low pass LOESS. 0 (constant) or 1 (constant and trend).
        robust: bool, Flag indicating whether to use a weighted version that is robust to some forms of outliers.
        seasonal_jump: int, Positive integer determining the linear interpolation step.
                       If larger than 1, the LOESS is used every seasonal_jump points and linear interpolation is between fitted points.
                       Higher values reduce estimation time.
        trend_jump: int, Positive integer determining the linear interpolation step.
                    If larger than 1, the LOESS is used every trend_jump points and values between the two are linearly interpolated.
                    Higher values reduce estimation time.
        low_pass_jump: int, Positive integer determining the linear interpolation step.
                       If larger than 1, the LOESS is used every low_pass_jump points and values between the two are linearly interpolated.
                       Higher values reduce estimation time.
    """

    def __init__(
        self,
        method: str,
        m: int,
        method_params: Optional[Params] = None,
        decomposition: str = "multiplicative",
        seasonal: int = 7,
        trend: Optional[int] = None,
        low_pass: Optional[int] = None,
        seasonal_deg: int = 1,
        trend_deg: int = 1,
        low_pass_deg: int = 1,
        robust: bool = False,
        seasonal_jump: int = 1,
        trend_jump: int = 1,
        low_pass_jump: int = 1,
    ) -> None:
        super().__init__()
        self.method = method
        self.m = m
        self.method_params = method_params
        self.decomposition = decomposition
        self.seasonal = seasonal
        self.trend = trend
        self.low_pass = low_pass
        self.seasonal_deg = seasonal_deg
        self.trend_deg = trend_deg
        self.low_pass_deg = low_pass_deg
        self.robust = robust
        self.seasonal_jump = seasonal_jump
        self.trend_jump = trend_jump
        self.low_pass_jump = low_pass_jump
        self.validate_params()
        logging.debug("Initialized STFLParams instance.")

    def validate_params(self) -> None:
        """Validate the parameters for STLF model"""

        if self.method not in MODELS:
            msg = f"Only support {', '.join(MODELS)} method, but get {self.method}."
            logging.error(msg)
            raise ValueError(msg)

        if self.method_params is None:
            if self.method == "theta":
                self.method_params = theta.ThetaParams(m=1)
            elif self.method == "linear":
                self.method_params = linear_model.LinearModelParams()
            elif self.method == "simple":
                self.method_params = simple_heuristic_model.SimpleHeuristicModelParams()
            else:
                assert self.method == "quadratic"
                self.method_params = quadratic_model.QuadraticModelParams()

            if self.decomposition not in ["additive", "multiplicative"]:
                msg = "decomposition can be only `additive` or `multiplicative` strings"
                logging.error(msg)
                raise ValueError(msg)


class STLFModel(Model[STLFParams]):
    """Model class for STLF

    This class provides fit, predict, and plot methods for STLF model

    Attributes:
        data: the input time series data as :class:`kats.consts.TimeSeriesData`
        params: the parameter class defined with `STLFParams`
    """

    decomp: Optional[Dict[str, TimeSeriesData]] = None
    sea_data: Optional[TimeSeriesData] = None
    trend_data: Optional[TimeSeriesData] = None
    desea_data: Optional[TimeSeriesData] = None
    model: Optional[Union[LinearModel, QuadraticModel, SimpleHeuristicModel, ThetaModel]] = None
    freq: Optional[str] = None
    alpha: Optional[float] = None
    y_fcst: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]] = None
    fcst_lower: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]] = None
    fcst_upper: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]] = None
    dates: Optional[pd.DatetimeIndex] = None
    fcst_df: Optional[pd.DataFrame] = None
    deseasonal_operator: Callable(Union[_operator.truediv, _operator.sub])[
        [
            Union[pd.Series, pd.DataFrame],
            Union[pd.Series, pd.DataFrame],
        ],
        Union[pd.Series, pd.DataFrame],
    ]
    reseasonal_operator: Callable(Union[_operator.mul, _operator.add])[
        [
            Union[pd.Series, pd.DataFrame],
            Union[pd.Series, pd.DataFrame],
        ],
        Union[np.ndarray, pd.Series, pd.DataFrame],
    ]

    def __init__(self, data: TimeSeriesData, params: STLFParams) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)
        self.n: int = self.data.value.shape[0]

        if self.params.m > self.n:
            msg = "The seasonality length m must be smaller than the length of time series"
            logging.error(msg)
            raise ValueError(msg)

        if self.params.decomposition == "multiplicative":
            self.deseasonal_operator = operator.truediv
            self.reseasonal_operator = operator.mul
        else:
            assert self.params.decomposition == "additive"
            self.deseasonal_operator = operator.sub
            self.reseasonal_operator = operator.add

    def deseasonalize(self) -> STLFModel:
        """De-seasonalize the time series data

        Args:
            None

        Returns:
            This object, with `decomp`, `sea_data`, and `desea_data` attributes
            set to the decomposition results.
        """

        # create decomposer for time series decomposition
        decomposer = TimeSeriesDecomposition(
            data=cast(TimeSeriesData, self.data),
            decomposition=self.params.decomposition,
            method="STL",
            period=self.params.m,
            seasonal=self.params.seasonal,
            trend=self.params.trend,
            low_pass=self.params.low_pass,
            seasonal_deg=self.params.seasonal_deg,
            trend_deg=self.params.trend_deg,
            low_pass_deg=self.params.low_pass_deg,
            robust=self.params.robust,
            seasonal_jump=self.params.seasonal_jump,
            trend_jump=self.params.trend_jump,
            low_pass_jump=self.params.low_pass_jump,
        )

        self.decomp = decomp = decomposer.decomposer()

        self.sea_data = copy(decomp["seasonal"])
        self.trend_data = copy(decomp["trend"])
        self.desea_data = desea_data = copy(self.data)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        desea_data.value = self.deseasonal_operator(desea_data.value, decomp["seasonal"].value)

        return self

    # pyre-fixme[15]: `fit` overrides method defined in `Model` inconsistently.
    def fit(self, **kwargs: Any) -> STLFModel:
        """Fit STLF model

        Args:
            None

        Returns:
            The fitted STLF model object
        """

        logging.debug("Call fit() with parameters. " "kwargs:{kwargs}".format(kwargs=kwargs))
        self.deseasonalize()
        if self.params.method == "simple":
            data = self.trend_data
        else:
            data = self.desea_data
        assert data is not None
        if self.params.method == "theta":
            model = theta.ThetaModel(data=data, params=cast(ThetaParams, self.params.method_params))
        elif self.params.method == "linear":
            model = linear_model.LinearModel(data=data, params=cast(LinearModelParams, self.params.method_params))
        elif self.params.method == "simple":
            model = simple_heuristic_model.SimpleHeuristicModel(
                data=data,
                params=cast(SimpleHeuristicModelParams, self.params.method_params),
            )
        else:
            assert self.params.method == "quadratic"
            model = quadratic_model.QuadraticModel(
                data=data, params=cast(QuadraticModelParams, self.params.method_params)
            )
        model.fit()
        self.model = model
        return self

    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, *args: Any, include_history: bool = False, **kwargs: Any) -> pd.DataFrame:
        """predict with the fitted STLF model

        Args:
            steps: the steps or length of prediction horizon
            include_history: if include the historical data, default as False

        Returns:
            The predicted dataframe with following columns:
            `time`, `fcst`, `fcst_lower`, and `fcst_upper`
        """
        model = self.model
        if model is None:
            raise ValueError("Call fit() before predict().")
        decomp = self.decomp
        assert decomp is not None

        logging.debug(
            "Call predict() with parameters. " "steps:{steps}, kwargs:{kwargs}".format(steps=steps, kwargs=kwargs)
        )
        self.include_history = include_history
        # pyre-fixme[16]: `Optional` has no attribute `time`.
        self.freq = kwargs.get("freq", pd.infer_freq(self.data.time))
        if self.freq is None:
            logging.info("Could not infer freq, will default to 'D' (daily)")
        self.alpha = kwargs.get("alpha", 0.05)

        # trend forecast
        fcst = model.predict(steps=steps, include_history=include_history)

        # re-seasonalize
        m = self.params.m
        rep = math.trunc(1 + fcst.shape[0] / m)

        seasonality = decomp["seasonal"].value[-m:]

        self.y_fcst = self.reseasonal_operator(fcst.fcst, np.tile(seasonality, rep)[: fcst.shape[0]])
        if ("fcst_lower" in fcst.columns) and ("fcst_upper" in fcst.columns):
            self.fcst_lower = self.reseasonal_operator(fcst.fcst_lower, np.tile(seasonality, rep)[: fcst.shape[0]])
            self.fcst_upper = self.reseasonal_operator(fcst.fcst_upper, np.tile(seasonality, rep)[: fcst.shape[0]])
        logging.info("Generated forecast data from STLF model.")
        logging.debug("Forecast data: {fcst}".format(fcst=self.y_fcst))

        # TODO: create empirical uncertainty interval
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)
        self.dates = dates[dates != last_date]  # Return correct number of periods

        if include_history:
            self.dates = np.concatenate((pd.to_datetime(self.data.time), self.dates))

        self.fcst_df = fcst_df = pd.DataFrame(
            {
                "time": self.dates,
                "fcst": self.y_fcst,
                "fcst_lower": self.fcst_lower,
                "fcst_upper": self.fcst_upper,
            },
            copy=False,
        )

        logging.debug("Return forecast data: {fcst_df}".format(fcst_df=fcst_df))
        return fcst_df

    def __str__(self) -> str:
        """AR net moddel as a string

        Args:
            None

        Returns:
            String representation of the model name
        """

        return "STLF"

    @staticmethod
    # pyre-fixme[15]: `get_parameter_search_space` overrides method defined in
    #  `Model` inconsistently.
    def get_parameter_search_space() -> List[Dict[str, Any]]:
        """Provide a parameter space for STLF model

        Move the implementation of get_parameter_search_space() out of stlf
        to keep HPT implementation tighter, and avoid the dependency conflict issue.

        Args:
            None

        Returns:
            List of dicts contains parameter search space
        """

        return get_default_stlf_parameter_search_space()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

import matplotlib as mpl
from pytest import Session


def pytest_sessionstart(session: Session) -> None:
    # Set the matplotlib backend to Agg for UI-less testing
    # unless the developer manually overrides by setting
    # MPLBACKEND to something else (such as "TkAgg").
    if "MPLBACKEND" not in os.environ:
        os.environ["MPLBACKEND"] = "agg"
        # The above should be enough, but I found I needed to use:
        mpl.use("agg")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""A module for meta-learner model selection.

This module contains:
    - :class:`MetaLearnModelSelect` for meta-learner models selection, which recommends the forecasting model based on time series or time series features;
    - :class:`RandomDownSampler` for creating balanced dataset via downsampling.
"""

import ast
import logging
from collections import Counter
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import joblib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn import metrics
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC

from kats.consts import TimeSeriesData
from kats.tsfeatures.tsfeatures import TsFeatures


class MetaLearnModelSelect:
    """Meta-learner framework on forecasting model selection.
    This framework uses classification algorithms to recommend suitable forecasting models.
    For training, it uses time series features as inputs and the best forecasting models as labels.
    For prediction, it takes time series or time series features as inputs to predict the most suitable forecasting model.
    The class provides count_category, preprocess, plot_feature_comparison, get_corr_mtx, plot_corr_heatmap, train, pred, pred_by_feature, pred_fuzzy, load_model and save_model.

    Attributes:
        metadata: Optional; A list of dictionaries representing the meta-data of time series.
                  Each dictionary d must contain at least 3 components: 'hpt_res', 'features' and 'best_model'. d['hpt_res'] represents the best hyper-parameters for each candidate model and the corresponding errors;
                  d['features'] are time series features, and d['best_model'] is a string representing the best candidate model of the corresponding time series data.
                  metadata should not be None unless load_model is True. Default is None.
        load_model: Optional; A boolean to specify whether or not to load a trained model. Default is False.

    Sample Usage:
        >>> mlms = MetaLearnModelSelect(data)
        >>> mlms.train(n_trees=200, test_size=0.1, eval_method='mean') # Train a meta-learner model selection model.
        >>> mlms.pred(TSdata) # Predict/recommend forecasting model for a new time series data.
        >>> mlms2.pred(TSdata, n_top=3) # Predict/recommend the top 3 most suitable forecasting model.
        >>> mlms.save_model("mlms.pkl") # Save the trained model.
        >>> mlms2 = MetaLearnModelSelect(metadata=None, load_model=True) # Create a new object and then load a pre-trained model.
        >>> mlms2.load_model("mlms.pkl")
    """

    def __init__(self, metadata: Optional[List[Dict[str, Any]]] = None, load_model: bool = False) -> None:
        if not load_model:
            # pyre-fixme[6]: Expected `Sized` for 1st param but got
            #  `Optional[List[typing.Any]]`.
            if len(metadata) <= 30:
                msg = "Dataset is too small to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            if metadata is None:
                msg = "Missing metadata!"
                logging.error(msg)
                raise ValueError(msg)

            if "hpt_res" not in metadata[0]:
                msg = "Missing best hyper-params, not able to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            if "features" not in metadata[0]:
                msg = "Missing time series features, not able to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            if "best_model" not in metadata[0]:
                msg = "Missing best models, not able to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            # pyre-fixme[4]: Attribute must be annotated.
            self.metadata = metadata
            self._reorganize_data()
            self._validate_data()

            # pyre-fixme[4]: Attribute must be annotated.
            self.scale = False
            # pyre-fixme[4]: Attribute must be annotated.
            self.clf = None
        elif load_model:
            pass
        else:
            msg = "Fail to initiate MetaLearnModelSelect."
            raise ValueError(msg)

    def _reorganize_data(self) -> None:
        hpt_list = []
        metadataX_list = []
        metadataY_list = []
        for i in range(len(self.metadata)):
            if isinstance(self.metadata[i]["hpt_res"], str):
                hpt_list.append(ast.literal_eval(self.metadata[i]["hpt_res"]))
            else:
                hpt_list.append(self.metadata[i]["hpt_res"])

            if isinstance(self.metadata[i]["features"], str):
                metadataX_list.append(list(ast.literal_eval(self.metadata[i]["features"]).values()))
            else:
                metadataX_list.append(list(self.metadata[i]["features"].values()))

            metadataY_list.append(self.metadata[i]["best_model"])

        self.col_namesX = list(self.metadata[0]["features"].keys())
        self.hpt = pd.Series(hpt_list, name="hpt", copy=False)
        self.metadataX = pd.DataFrame(metadataX_list, columns=self.col_namesX, copy=False)
        self.metadataX.fillna(0, inplace=True)
        self.metadataY = pd.Series(metadataY_list, name="y", copy=False)
        self.x_mean = np.average(self.metadataX.values, axis=0)
        self.x_std = np.std(self.metadataX.values, axis=0)
        self.x_std[self.x_std == 0] = 1.0

    # pyre-fixme[3]: Return type must be annotated.
    def _validate_data(self):
        num_class = self.metadataY.nunique()
        if num_class == 1:
            msg = "Only one class in the label column (best_model), not able to train a classifier!"
            logging.error(msg)
            raise ValueError(msg)

        local_count = list(self.count_category().values())
        if min(local_count) * num_class < 30:
            msg = "Not recommend to do downsampling! Dataset will be too small after downsampling!"
            logging.info(msg)
        elif max(local_count) > min(local_count) * 5:
            msg = (
                "Number of obs in majority class is much greater than in minority class. Downsampling is recommended!"
            )
            logging.info(msg)
        else:
            msg = "No significant data imbalance problem, no need to do downsampling."
            logging.info(msg)

    def count_category(self) -> Dict[str, int]:
        """Count the number of observations of each candidate model in meta-data.

        Returns:
            A dictionary storing the number of observations of each candidate model in meta-data.
        """

        return Counter(self.metadataY)

    def preprocess(self, downsample: bool = True, scale: bool = False) -> None:
        """Pre-process meta data before training a classifier.

        There are 2 options in this function: 1) whether or not to downsample meta-data to ensure each candidate model has the same number of observations;
        and 2) whether or not to rescale the time series features to zero-mean and unit-variance.

        Args:
            downsample: Optional; A boolean to specify whether or not to downsample meta-data to ensure each candidate model has the same number of observations.
                        Default is True.
            scale: Optional; A boolean to specify whether or not to rescale the time series features to zero-mean and unit-variance.

        Returns:
            None
        """

        if downsample:
            self.hpt, self.metadataX, self.metadataY = RandomDownSampler(
                self.hpt, self.metadataX, self.metadataY
            ).fit_resample()
            logging.info("Successfully applied random downsampling!")
            self.x_mean = np.average(self.metadataX.values, axis=0)
            self.x_std = np.std(self.metadataX.values, axis=0)
            self.x_std[self.x_std == 0] = 1.0

        if scale:
            self.scale = True
            self.metadataX = (self.metadataX - self.x_mean) / self.x_std
            logging.info(
                "Successfully scaled data by centering to the mean and component-wise scaling to unit variance!"
            )

    def plot_feature_comparison(
        self,
        i: int,
        j: int,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[int, int]] = None,
    ) -> plt.Axes:
        """Generate the time series features comparison plot.

        Args:
            i: A integer representing the index of one feature vector from feature matrix to be compared.
            j: A integer representing the other index of one feature vector from feature matrix to be compared.
            ax: optional axes to use for plotting.
            figsize: optional figure size to create. If None, use (12, 6).

        Returns:
            The matplotlib Axes.
        """

        combined = pd.concat([self.metadataX.iloc[i], self.metadataX.iloc[j]], axis=1, copy=False)
        combined.columns = [
            f"{self.metadataY.iloc[i]} model",
            f"{self.metadataY.iloc[j]} model",
        ]
        if figsize is None:
            figsize = (12, 6)
        # pyre-fixme[29]: `CachedAccessor` is not a function.
        return combined.plot(kind="bar", figsize=figsize, ax=ax)

    def get_corr_mtx(self) -> pd.DataFrame:
        """Calculate correlation matrix of feature matrix.

        Returns:
            A pd.DataFrame representing the correlation matrix of time series features.
        """

        return self.metadataX.corr()

    def plot_corr_heatmap(
        self,
        camp: str = "RdBu_r",
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[int, int]] = None,
    ) -> plt.Axes:
        """Generate heat-map for correlation matrix of feature matrix.

        Args:
            cmap: Optional; A string representing the color bar used to generate heat-map. Default is "RdBu_r".
            ax: optional axes to use for plotting.
            figsize: optional figure size to create. If None, use (8, 6).

        Returns:
            The matplotlib Axes.
        """
        if ax is None:
            if figsize is None:
                figsize = (8, 6)
            _, ax = plt.subplots(figsize=figsize)

        return sns.heatmap(
            self.get_corr_mtx(),
            cmap=camp,
            yticklabels=self.metadataX.columns,
            xticklabels=self.metadataX.columns,
            ax=ax,
        )

    def train(
        self,
        method: str = "RandomForest",
        eval_method: str = "mean",
        test_size: float = 0.1,
        n_trees: int = 500,
        n_neighbors: int = 5,
    ) -> Dict[str, Any]:
        """Train a meta-learner model selection model (i.e., a classifier).

        Args:
            method: Optional; A string representing the name of the classification algorithm. Can be 'RandomForest', 'GBDT', 'SVM', 'KNN' or 'NaiveBayes'. Default is 'RandomForest'.
            eval_method: Optional; A string representing the aggregation method used for computing errors. Can be 'mean' or 'median'. Default is 'mean'.
            test_size: Optional; A float representing the proportion of test set, which should be within (0, 1). Default is 0.1.
            n_trees: Optional; An integer representing the number of trees in random forest model. Default is 500.
            n_neighbors: Optional; An integer representing the number of neighbors in KNN model. Default is 5.

        Returns:
            A dictionary summarizing the performance of the trained classifier on both training and validation set.
        """

        if method not in ["RandomForest", "GBDT", "SVM", "KNN", "NaiveBayes"]:
            msg = "Only support RandomForest, GBDT, SVM, KNN, and NaiveBayes method."
            logging.error(msg)
            raise ValueError(msg)

        if eval_method not in ["mean", "median"]:
            msg = "Only support mean and median as evaluation method."
            logging.error(msg)
            raise ValueError(msg)

        if test_size <= 0 or test_size >= 1:
            msg = "Illegal test set."
            logging.error(msg)
            raise ValueError(msg)

        x_train, x_test, y_train, y_test, hpt_train, hpt_test = train_test_split(
            self.metadataX, self.metadataY, self.hpt, test_size=test_size
        )

        if method == "RandomForest":
            clf = RandomForestClassifier(n_estimators=n_trees)
        elif method == "GBDT":
            clf = GradientBoostingClassifier()
        elif method == "SVM":
            clf = make_pipeline(StandardScaler(), SVC(gamma="auto"))
        elif method == "KNN":
            clf = KNeighborsClassifier(n_neighbors=n_neighbors)
        else:
            clf = GaussianNB()

        clf.fit(x_train, y_train)
        y_fit = clf.predict(x_train)
        y_pred = clf.predict(x_test)

        # calculate model errors
        fit_error, pred_error = {}, {}

        # evaluate method
        em = np.mean if eval_method == "mean" else np.median

        # meta learning errors
        fit_error["meta-learn"] = em([hpt_train.iloc[i][c][-1] for i, c in enumerate(y_fit)])
        pred_error["meta-learn"] = em([hpt_test.iloc[i][c][-1] for i, c in enumerate(y_pred)])

        # pre-selected model errors, for all candidate models
        for label in self.metadataY.unique():
            fit_error[label] = em([hpt_train.iloc[i][label][-1] for i in range(len(hpt_train))])
            pred_error[label] = em([hpt_test.iloc[i][label][-1] for i in range(len(hpt_test))])

        self.clf = clf
        return {
            "fit_error": fit_error,
            "pred_error": pred_error,
            "clf_accuracy": metrics.accuracy_score(y_test, y_pred),
        }

    def save_model(self, file_name: str) -> None:
        """Save the trained model.

        Args:
            file_name: A string representing the path to save the trained model.

        Returns:
            None.
        """

        if self.clf is None:
            msg = "Haven't trained a model."
            logging.error(msg)
            raise ValueError(msg)
        else:
            joblib.dump(self.__dict__, file_name)
            logging.info("Successfully saved the trained model!")

    def load_model(self, file_name: str) -> None:
        """Load a pre-trained model.

        Args:
            file_name: A string representing the path to load the pre-trained model.

        Returns:
            None.
        """

        try:
            self.__dict__ = joblib.load(file_name)
            logging.info("Successfully loaded a pre-trained model!")
        except Exception:
            msg = "No existing pre-trained model. Please change file path or train a model first!"
            logging.error(msg)
            raise ValueError(msg)

    def pred(self, source_ts: TimeSeriesData, ts_scale: bool = True, n_top: int = 1) -> Union[str, List[str]]:
        """Predict the best forecasting model for a new time series data.

        Args:
            source_ts: :class:`kats.consts.TimeSeriesData` object representing the new time series data.
            ts_scale: Optional; A boolean to specify whether or not to rescale time series data (i.e., normalizing it with its maximum vlaue) before calculating features. Default is True.
            n_top: Optional; A integer for the number of top model names to return. Default is 1.

        Returns:
            A string or a list of strings of the names of forecasting models.
        """

        ts = TimeSeriesData(pd.DataFrame(source_ts.to_dataframe().copy()))
        if self.clf is None:
            msg = "Haven't trained a model. Please train a model or load a model before predicting."
            logging.error(msg)
            raise ValueError(msg)

        if ts_scale:
            # scale time series to make ts features more stable
            ts.value /= ts.value.max()
            msg = "Successful scaled! Each value of TS has been divided by the max value of TS."
            logging.info(msg)

        new_features = TsFeatures().transform(ts)
        # pyre-fixme[16]: `List` has no attribute `values`.
        new_features_vector = np.asarray(list(new_features.values()))
        if np.any(np.isnan(new_features_vector)):
            msg = (
                "Features of the test time series contains NaN value, consider processing it. Features are: "
                f"{new_features}. Fill in NaNs with 0."
            )
            logging.warning(msg)
        return self.pred_by_feature([new_features_vector], n_top=n_top)[0]

    def pred_by_feature(
        self,
        source_x: Union[np.ndarray, List[np.ndarray], pd.DataFrame],
        n_top: int = 1,
    ) -> np.ndarray:
        """Predict the best forecasting models given a list/dataframe of time series features
        Args:
            source_x: the time series features of the time series that one wants to predict, can be a np.ndarray, a list of np.ndarray or a pd.DataFrame.
            n_top: Optional; An integer for the number of top model names to return. Default is 1.

        Returns:
            An array of strings representing the forecasing models. If n_top=1, a 1-d np.ndarray will be returned. Otherwise, a 2-d np.ndarray will be returned.
        """

        if self.clf is None:
            msg = "Haven't trained a model. Please train a model or load a model before predicting."
            logging.error(msg)
            raise ValueError(msg)
        if isinstance(source_x, List):
            x = np.row_stack(source_x)
        elif isinstance(source_x, np.ndarray):
            x = source_x.copy()
        else:
            msg = f"Invalid source_x type: {type(source_x)}."
            logging.error(msg)
            raise ValueError(msg)
        if self.scale:
            x = (x - self.x_mean) / self.x_std
        x[np.isnan(x)] = 0.0
        if n_top == 1:
            return self.clf.predict(x)
        prob = self.clf.predict_proba(x)
        order = np.argsort(-prob, axis=1)
        classes = np.array(self.clf.classes_)
        return classes[order][:, :n_top]

    def _bootstrap(self, data: np.ndarray, rep: int = 200) -> float:
        """Helper function for bootstrap test and returns the pvalue."""

        diff = data[:, 0] - data[:, 1]
        n = len(diff)
        idx = np.random.choice(np.arange(n), n * rep)
        sample = diff[idx].reshape(-1, n)
        bs = np.average(sample, axis=1)
        pvalue = np.average(bs < 0)
        return pvalue

    def pred_fuzzy(self, source_ts: TimeSeriesData, ts_scale: bool = True, sig_level: float = 0.2) -> Dict[str, Any]:
        """Predict a forecasting model for a new time series data using fuzzy method.

        The fuzzy method returns the best candiate model and the second best model will be returned if there is no statistically significant difference between them.
        The statistical test is based on the bootstrapping samples drawn from the fitted random forest model. This function is only available for random forest classifier.

        Args:
            source_ts: :class:`kats.consts.TimeSeriesData` object representing the new time series data.
            ts_scale: Optional; A boolean to specify whether or not to rescale time series data (i.e., normalizing it with its maximum vlaue) before calculating features. Default is True.
            sig_level: Optional; A float representing the significance level for bootstrap test. If pvalue>=sig_level, then we deem there is no difference between the best and the second best model.
                       Default is 0.2.

        Returns:
            A dictionary of prediction results, including forecasting models, their probability of being th best forecasting models and the pvalues of bootstrap tests.
        """

        ts = TimeSeriesData(pd.DataFrame(source_ts.to_dataframe().copy()))
        if ts_scale:
            # scale time series to make ts features more stable
            ts.value /= ts.value.max()
        # pyre-fixme[16]: `List` has no attribute `values`.
        test = np.asarray(list(TsFeatures().transform(ts).values()))
        test[np.isnan(test)] = 0.0
        if self.scale:
            test = (test - self.x_mean) / self.x_std
        test = test.reshape([1, -1])
        m = len(self.clf.estimators_)
        data = np.array([self.clf.estimators_[i].predict_proba(test)[0] for i in range(m)])
        prob = self.clf.predict_proba(test)[0]
        idx = np.argsort(-prob)[:2]
        pvalue = self._bootstrap(data[:, idx[:2]])
        if pvalue >= sig_level:
            label = self.clf.classes_[idx[:2]]
            prob = prob[idx[:2]]
        else:
            label = self.clf.classes_[idx[:1]]
            prob = prob[idx[:1]]
        ans = {"label": label, "probability": prob, "pvalue": pvalue}
        return ans

    # pyre-fixme[3]: Return type must be annotated.
    def __str__(self):
        return "MetaLearnModelSelect"


class RandomDownSampler:
    """An assistant class for class MetaLearnModelSelect to do random downsampling.

    RandomDownSampler provides methods for creating a balanced dataset via downsampling. It contains fit_resample.

    Attributes:
        hpt: A `pandas.Series` object storing the best hyper-parameters and the corresponding errors for each model.
        dataX: A `pandas.DataFrame` object representing the time series features matrix.
        dataY: A `pandas.Series` object representing the best models for the corresponding time series.
    """

    def __init__(self, hpt: pd.Series, dataX: pd.DataFrame, dataY: pd.Series) -> None:
        self.hpt = hpt
        self.dataX = dataX
        self.dataY = dataY
        # pyre-fixme[4]: Attribute must be annotated.
        self.col_namesX = self.dataX.columns

    def fit_resample(self) -> Tuple[pd.Series, pd.DataFrame, pd.Series]:
        """Create balanced dataset via random downsampling.

        Returns:
            A tuple containing the `pandas.Series` object of the best hyper-parameters and the corresponding errors, the `pandas.DataFrame` object of the downsampled time series features,
            and the `pandas.Series` object of the downsampled best models for the corresponding time series.
        """

        resampled_x, resampled_y, resampled_hpt = [], [], []
        # naive down-sampler technique for data imbalance problem
        min_n = min(Counter(self.dataY).values())

        idx_dict = defaultdict(list)
        for i, c in enumerate(self.dataY):
            idx_dict[c].append(i)

        for key in idx_dict:
            idx_dict[key] = np.random.choice(idx_dict[key], size=min_n, replace=False)
            resampled_x += self.dataX.iloc[np.asarray(idx_dict[key]), :].values.tolist()
            resampled_y += list(self.dataY.iloc[np.asarray(idx_dict[key])])
            resampled_hpt += list(self.hpt.iloc[np.asarray(idx_dict[key])])

        resampled_x = pd.DataFrame(resampled_x, copy=False)
        resampled_x.columns = self.col_namesX

        resampled_y = pd.Series(resampled_y, name="y", copy=False)
        resampled_hpt = pd.Series(resampled_hpt, name="hpt", copy=False)

        return resampled_hpt, resampled_x, resampled_y



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kats',
 'kats.compat',
 'kats.data',
 'kats.detectors',
 'kats.detectors.meta_learning',
 'kats.metrics',
 'kats.models',
 'kats.models.ensemble',
 'kats.models.metalearner',
 'kats.models.nowcasting',
 'kats.models.reconciliation',
 'kats.tests',
 'kats.tests.compat',
 'kats.tests.data',
 'kats.tests.detectors',
 'kats.tests.detectors.meta_learning',
 'kats.tests.metrics',
 'kats.tests.models',
 'kats.tests.tsfeatures',
 'kats.tests.utils',
 'kats.tsfeatures',
 'kats.utils']

package_data = \
{'': ['*'], 'kats.tests': ['baseline/*'], 'kats.tests.models': ['baseline/*']}

install_requires = \
['matplotlib==3.6.0',
 'numba==0.56.3',
 'numpy==1.21.0',
 'pandas==1.3.5',
 'python-dateutil==2.8.0',
 'scikit-learn==1.1.2',
 'scipy==1.7.3',
 'seaborn==0.11.1',
 'statsmodels==0.12.2',
 'torch==1.12.1']

setup_kwargs = {
    'name': 'tinykats-corestack',
    'version': '0.2.0',
    'description': 'Stripped down version of Kats (facebookresearch/Kats)',
    'long_description': None,
    'author': 'Hem Chandra Padhalni',
    'author_email': 'hem.chandra@corestack.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any
from typing import Optional
from typing import Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import norm  # @manual
from scipy.stats import zscore

from kats.consts import TimeSeriesChangePoint
from kats.consts import TimeSeriesData
from kats.detectors.detector import Detector


class RobustStatChangePoint(TimeSeriesChangePoint):
    def __init__(
        self,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        confidence: float,
        index: int,
        metric: float,
    ) -> None:
        super().__init__(start_time, end_time, confidence)
        self._metric = metric
        self._index = index

    @property
    def metric(self) -> float:
        return self._metric

    @property
    def index(self) -> int:
        return self._index


class RobustStatDetector(Detector):
    def __init__(self, data: TimeSeriesData) -> None:
        super(RobustStatDetector, self).__init__(data=data)
        if not self.data.is_univariate():
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)
        self.zscore: Optional[np.ndarray] = None

    # pyre-fixme[14]: `detector` overrides method defined in `Detector` inconsistently.
    def detector(
        self,
        p_value_cutoff: float = 1e-2,
        smoothing_window_size: int = 5,
        comparison_window: int = -2,
    ) -> Sequence[RobustStatChangePoint]:
        time_col_name = self.data.time.name
        val_col_name = self.data.value.name

        data_df = self.data.to_dataframe()
        data_df = data_df.set_index(time_col_name)

        df_ = data_df.loc[:, val_col_name].rolling(window=smoothing_window_size)
        df_ = (
            # Smooth
            df_.mean()
            .fillna(method="bfill")
            # Make spikes standout
            .diff(comparison_window)
            .fillna(0)
        )

        y_zscores = zscore(df_)
        self.zscore = y_zscores
        p_values = norm.sf(np.abs(y_zscores))
        ind = np.where(p_values < p_value_cutoff)[0]

        if len(ind) == 0:
            return []  # empty list for no change points

        change_points = []

        prev_idx = -1
        for idx in ind:
            if prev_idx != -1 and (idx - prev_idx) < smoothing_window_size:
                continue

            prev_idx = idx
            cp = RobustStatChangePoint(
                start_time=data_df.index.values[idx],
                end_time=data_df.index.values[idx],
                # pyre-fixme[16]: `float` has no attribute `__getitem__`.
                confidence=1 - p_values[idx],
                index=idx,
                metric=float(df_.iloc[idx]),
            )

            change_points.append(cp)

        return change_points

    def plot(self, change_points: Sequence[RobustStatChangePoint], **kwargs: Any) -> plt.Axes:
        time_col_name = self.data.time.name
        val_col_name = self.data.value.name

        data_df = self.data.to_dataframe()

        _, ax = plt.subplots()
        ax.plot(data_df[time_col_name].to_numpy(), data_df[val_col_name].to_numpy())

        changepoint_annotated = False
        for change in change_points:
            ax.axvline(x=change.start_time, color="red")
            changepoint_annotated = True
        if not changepoint_annotated:
            logging.warning("No change points detected!")

        return ax
